# -*- coding: utf-8 -*-

"""OBO Readers."""

import logging
from datetime import datetime
from pathlib import Path
from typing import Any, Iterable, List, Mapping, Optional, Tuple, Union

import bioregistry
import networkx as nx
from more_itertools import pairwise
from tqdm.auto import tqdm

from .constants import DATE_FORMAT, PROVENANCE_PREFIXES
from .identifier_utils import MissingPrefix, normalize_curie
from .registries import curie_has_blacklisted_prefix, curie_is_blacklisted, remap_prefix
from .struct import (
    Obo,
    Reference,
    Synonym,
    SynonymSpecificities,
    SynonymSpecificity,
    SynonymTypeDef,
    Term,
    TypeDef,
    make_ad_hoc_ontology,
)
from .struct.struct import DEFAULT_SYNONYM_TYPE
from .struct.typedef import default_typedefs, develops_from, has_part, part_of
from .utils.misc import cleanup_version

__all__ = [
    "from_obo_path",
    "from_obonet",
]

logger = logging.getLogger(__name__)

# FIXME use bioontologies
# RELATION_REMAPPINGS: Mapping[str, Tuple[str, str]] = bioontologies.upgrade.load()
RELATION_REMAPPINGS: Mapping[str, Tuple[str, str]] = {
    "part_of": part_of.pair,
    "has_part": has_part.pair,
    "develops_from": develops_from.pair,
    "seeAlso": ("rdf", "seeAlso"),
    "dc-contributor": ("dc", "contributor"),
    "dc-creator": ("dc", "creator"),
}


def from_obo_path(
    path: Union[str, Path], prefix: Optional[str] = None, *, strict: bool = True
) -> Obo:
    """Get the OBO graph from a path."""
    import obonet

    logger.info("[%s] parsing with obonet from %s", prefix or "", path)
    with open(path) as file:
        graph = obonet.read_obo(
            tqdm(
                file,
                unit_scale=True,
                desc=f'[{prefix or ""}] parsing obo',
                disable=None,
                leave=False,
            )
        )

    if prefix:
        # Make sure the graph is named properly
        _clean_graph_ontology(graph, prefix)

    # Convert to an Obo instance and return
    return from_obonet(graph, strict=strict)


def from_obonet(graph: nx.MultiDiGraph, *, strict: bool = True) -> "Obo":  # noqa:C901
    """Get all of the terms from a OBO graph."""
    _ontology = graph.graph["ontology"]
    ontology = bioregistry.normalize_prefix(_ontology)  # probably always okay
    if ontology is None:
        raise ValueError(f"unknown prefix: {_ontology}")
    logger.info("[%s] extracting OBO using obonet", ontology)

    date = _get_date(graph=graph, ontology=ontology)
    name = _get_name(graph=graph, ontology=ontology)

    data_version = graph.graph.get("data-version")
    if not data_version:
        if date is not None:
            data_version = date.strftime("%Y-%m-%d")
            logger.info(
                "[%s] does not report a version. falling back to date: %s",
                ontology,
                data_version,
            )
        else:
            logger.warning("[%s] does not report a version nor a date", ontology)
    else:
        data_version = cleanup_version(data_version=data_version, prefix=ontology)
        if data_version is not None:
            logger.info("[%s] using version %s", ontology, data_version)
        elif date is not None:
            logger.info(
                "[%s] unrecognized version format, falling back to date: %s",
                ontology,
                data_version,
            )
            data_version = date.strftime("%Y-%m-%d")
        else:
            logger.warning(
                "[%s] UNRECOGNIZED VERSION FORMAT AND MISSING DATE: %s", ontology, data_version
            )

    if data_version and "/" in data_version:
        raise ValueError(f"[{ontology}] will not accept slash in data version: {data_version}")

    #: Parsed CURIEs to references (even external ones)
    reference_it = (
        Reference(
            prefix=prefix,
            identifier=bioregistry.standardize_identifier(prefix, identifier),
            # if name isn't available, it means its external to this ontology
            name=data.get("name"),
        )
        for prefix, identifier, data in _iter_obo_graph(graph=graph, strict=strict)
    )
    references: Mapping[Tuple[str, str], Reference] = {
        reference.pair: reference for reference in reference_it
    }

    #: CURIEs to typedefs
    typedefs: Mapping[Tuple[str, str], TypeDef] = {
        typedef.pair: typedef for typedef in iterate_graph_typedefs(graph, ontology)
    }

    synonym_typedefs: Mapping[str, SynonymTypeDef] = {
        synonym_typedef.curie: synonym_typedef
        for synonym_typedef in iterate_graph_synonym_typedefs(graph, ontology=ontology)
    }

    missing_typedefs = set()
    terms = []
    n_alt_ids, n_parents, n_synonyms, n_relations, n_properties, n_xrefs = 0, 0, 0, 0, 0, 0
    for prefix, identifier, data in _iter_obo_graph(graph=graph, strict=strict):
        if prefix != ontology or not data:
            continue

        identifier = bioregistry.standardize_identifier(prefix, identifier)
        reference = references[ontology, identifier]

        try:
            node_xrefs = list(iterate_node_xrefs(prefix=prefix, data=data, strict=strict))
        except MissingPrefix as e:
            e.reference = reference
            raise e
        xrefs, provenance = [], []
        for node_xref in node_xrefs:
            if node_xref.prefix in PROVENANCE_PREFIXES:
                provenance.append(node_xref)
            else:
                xrefs.append(node_xref)
        n_xrefs += len(xrefs)

        definition, definition_references = get_definition(
            data, prefix=prefix, identifier=identifier
        )
        if definition_references:
            provenance.extend(definition_references)

        try:
            alt_ids = list(iterate_node_alt_ids(data, strict=strict))
        except MissingPrefix as e:
            e.reference = reference
            raise e
        n_alt_ids += len(alt_ids)

        try:
            parents = list(
                iterate_node_parents(
                    data,
                    prefix=prefix,
                    identifier=identifier,
                    strict=strict,
                )
            )
        except MissingPrefix as e:
            e.reference = reference
            raise e
        n_parents += len(parents)

        synonyms = list(
            iterate_node_synonyms(
                data,
                synonym_typedefs,
                prefix=prefix,
                identifier=identifier,
                strict=strict,
            )
        )
        n_synonyms += len(synonyms)

        term = Term(
            reference=reference,
            definition=definition,
            parents=parents,
            synonyms=synonyms,
            xrefs=xrefs,
            provenance=provenance,
            alt_ids=alt_ids,
        )

        try:
            relations_references = list(
                iterate_node_relationships(
                    data,
                    prefix=ontology,
                    identifier=identifier,
                    strict=strict,
                )
            )
        except MissingPrefix as e:
            e.reference = reference
            raise e
        for relation, reference in relations_references:
            if (relation.prefix, relation.identifier) in typedefs:
                typedef = typedefs[relation.prefix, relation.identifier]
            elif (relation.prefix, relation.identifier) in default_typedefs:
                typedef = default_typedefs[relation.prefix, relation.identifier]
            else:
                if (relation.prefix, relation.identifier) not in missing_typedefs:
                    missing_typedefs.add((relation.prefix, relation.identifier))
                    logger.warning("[%s] has no typedef for %s", ontology, relation)
                    logger.debug("[%s] available typedefs: %s", ontology, set(typedefs))
                continue
            n_relations += 1
            term.append_relationship(typedef, reference)
        for prop, value in iterate_node_properties(data, term=term):
            n_properties += 1
            term.append_property(prop, value)
        terms.append(term)

    logger.info(
        f"[{ontology}] got {len(references):,} references, {len(typedefs):,} typedefs, {len(terms):,} terms,"
        f" {n_alt_ids:,} alt ids, {n_parents:,} parents, {n_synonyms:,} synonyms, {n_xrefs:,} xrefs,"
        f" {n_relations:,} relations, and {n_properties:,} properties",
    )

    return make_ad_hoc_ontology(
        _ontology=ontology,
        _name=name,
        _auto_generated_by=graph.graph.get("auto-generated-by"),
        _format_version=graph.graph.get("format-version"),
        _typedefs=list(typedefs.values()),
        _synonym_typedefs=list(synonym_typedefs.values()),
        _date=date,
        _data_version=data_version,
        terms=terms,
    )


def _clean_graph_ontology(graph, prefix: str) -> None:
    """Update the ontology entry in the graph's metadata, if necessary."""
    if "ontology" not in graph.graph:
        logger.warning('[%s] missing "ontology" key', prefix)
        graph.graph["ontology"] = prefix
    elif not graph.graph["ontology"].isalpha():
        logger.warning(
            "[%s] ontology=%s has a strange format. replacing with prefix",
            prefix,
            graph.graph["ontology"],
        )
        graph.graph["ontology"] = prefix


def _iter_obo_graph(
    graph: nx.MultiDiGraph,
    *,
    strict: bool = True,
) -> Iterable[Tuple[str, str, Mapping[str, Any]]]:
    """Iterate over the nodes in the graph with the prefix stripped (if it's there)."""
    for node, data in graph.nodes(data=True):
        prefix, identifier = normalize_curie(node, strict=strict)
        if prefix is None or identifier is None:
            continue
        yield prefix, identifier, data


def _get_date(graph, ontology: str) -> Optional[datetime]:
    try:
        rv = datetime.strptime(graph.graph["date"], DATE_FORMAT)
    except KeyError:
        logger.info("[%s] does not report a date", ontology)
        return None
    except ValueError:
        logger.info("[%s] reports a date that can't be parsed: %s", ontology, graph.graph["date"])
        return None
    else:
        return rv


def _get_name(graph, ontology: str) -> str:
    try:
        rv = graph.graph["name"]
    except KeyError:
        logger.info("[%s] does not report a name", ontology)
        rv = ontology
    return rv


def iterate_graph_synonym_typedefs(
    graph: nx.MultiDiGraph, *, ontology: str, strict: bool = False
) -> Iterable[SynonymTypeDef]:
    """Get synonym type definitions from an :mod:`obonet` graph."""
    for s in graph.graph.get("synonymtypedef", []):
        sid, name = s.split(" ", 1)
        name = name.strip().strip('"')
        if sid.startswith("http://") or sid.startswith("https://"):
            reference = Reference.from_iri(sid, name=name)
        elif ":" not in sid:  # assume it's ad-hoc
            reference = Reference(prefix=ontology, identifier=sid, name=name)
        else:  # assume it's a curie
            reference = Reference.from_curie(sid, name=name, strict=strict)

        if reference is None:
            if strict:
                raise ValueError(f"Could not parse {sid}")
            else:
                continue

        yield SynonymTypeDef(reference=reference)


def iterate_graph_typedefs(
    graph: nx.MultiDiGraph, default_prefix: str, *, strict: bool = True
) -> Iterable[TypeDef]:
    """Get type definitions from an :mod:`obonet` graph."""
    for typedef in graph.graph.get("typedefs", []):
        if "id" in typedef:
            curie = typedef["id"]
        elif "identifier" in typedef:
            curie = typedef["identifier"]
        else:
            raise KeyError

        name = typedef.get("name")
        if name is None:
            logger.debug("[%s] typedef %s is missing a name", graph.graph["ontology"], curie)

        if ":" in curie:
            reference = Reference.from_curie(curie, name=name, strict=strict)
        else:
            reference = Reference(prefix=graph.graph["ontology"], identifier=curie, name=name)
        if reference is None:
            logger.warning("[%s] unable to parse typedef CURIE %s", graph.graph["ontology"], curie)
            continue

        xrefs = []
        for curie in typedef.get("xref", []):
            _xref = Reference.from_curie(curie, strict=strict)
            if _xref:
                xrefs.append(_xref)
        yield TypeDef(reference=reference, xrefs=xrefs)


def get_definition(
    data, *, prefix: str, identifier: str
) -> Union[Tuple[None, None], Tuple[str, List[Reference]]]:
    definition = data.get("def")  # it's allowed not to have a definition
    if not definition:
        return None, None
    return _extract_definition(definition, prefix=prefix, identifier=identifier)


def _extract_definition(
    s: str,
    *,
    prefix: str,
    identifier: str,
    strict: bool = False,
) -> Union[Tuple[None, None], Tuple[str, List[Reference]]]:
    """Extract the definitions."""
    if not s.startswith('"'):
        raise ValueError("definition does not start with a quote")

    try:
        definition, rest = _quote_split(s)
    except ValueError:
        logger.warning("[%s:%s] could not parse definition: %s", prefix, identifier, s)
        return None, None

    if not rest.startswith("[") or not rest.endswith("]"):
        logger.warning("[%s:%s] problem with definition: %s", prefix, identifier, s)
        provenance = []
    else:
        provenance = _parse_trailing_ref_list(rest, strict=strict)
    return definition, provenance


def _get_first_nonquoted(s: str) -> Optional[int]:
    for i, (a, b) in enumerate(pairwise(s), start=1):
        if b == '"' and a != "\\":
            return i
    return None


def _quote_split(s: str) -> Tuple[str, str]:
    s = s.lstrip('"')
    i = _get_first_nonquoted(s)
    if i is None:
        raise ValueError
    return _clean_definition(s[:i].strip()), s[i + 1 :].strip()


def _clean_definition(s: str) -> str:
    # if '\t' in s:
    #     logger.warning('has tab')
    return (
        s.replace('\\"', '"').replace("\n", " ").replace("\t", " ").replace("\d", "")  # noqa:W605
    )


def _extract_synonym(
    s: str,
    synonym_typedefs: Mapping[str, SynonymTypeDef],
    *,
    prefix: str,
    identifier: str,
    strict: bool = True,
) -> Optional[Synonym]:
    # TODO check if the synonym is written like a CURIE... it shouldn't but I've seen it happen
    try:
        name, rest = _quote_split(s)
    except ValueError:
        logger.warning("[%s:%s] invalid synonym: %s", prefix, identifier, s)
        return None

    specificity: Optional[SynonymSpecificity] = None
    for _specificity in SynonymSpecificities:
        if rest.startswith(_specificity):
            specificity = _specificity
            rest = rest[len(_specificity) :].strip()
            break

    stype: Optional[SynonymTypeDef] = None
    if specificity is not None:  # go fishing for a synonym type definition
        for _stype in synonym_typedefs.values():
            # Since there aren't a lot of carefully defined synonym definitions, it
            # can appear as a string or curie. Therefore, we might see temporary prefixes
            # get added, so we should check against full curies as well as local unique
            # identifiers
            if rest.startswith(_stype.curie):
                rest = rest[len(_stype.curie) :].strip()
                stype = _stype
                break
            elif rest.startswith(_stype.identifier):
                rest = rest[len(_stype.identifier) :].strip()
                stype = _stype
                break

    if not rest.startswith("[") or not rest.endswith("]"):
        logger.warning("[%s:%s] problem with synonym: %s", prefix, identifier, s)
        return None

    provenance = _parse_trailing_ref_list(rest, strict=strict)
    return Synonym(
        name=name,
        specificity=specificity or "EXACT",
        type=stype or DEFAULT_SYNONYM_TYPE,
        provenance=provenance,
    )


def _parse_trailing_ref_list(rest, *, strict: bool = True):
    rest = rest.lstrip("[").rstrip("]")
    return [
        Reference.from_curie(curie.strip(), strict=strict)
        for curie in rest.split(",")
        if curie.strip()
    ]


def iterate_node_synonyms(
    data: Mapping[str, Any],
    synonym_typedefs: Mapping[str, SynonymTypeDef],
    *,
    prefix: str,
    identifier: str,
    strict: bool = False,
) -> Iterable[Synonym]:
    """Extract synonyms from a :mod:`obonet` node's data.

    Example strings:
    - "LTEC I" EXACT [Orphanet:93938,DOI:xxxx]
    - "LTEC I" EXACT [Orphanet:93938]
    - "LTEC I" [Orphanet:93938]
    - "LTEC I" []
    """
    for s in data.get("synonym", []):
        s = _extract_synonym(
            s, synonym_typedefs, prefix=prefix, identifier=identifier, strict=strict
        )
        if s is not None:
            yield s


HANDLED_PROPERTY_TYPES = {
    "xsd:string": str,
    "xsd:dateTime": datetime,
}


def iterate_node_properties(
    data: Mapping[str, Any], *, property_prefix: Optional[str] = None, term=None
) -> Iterable[Tuple[str, str]]:
    """Extract properties from a :mod:`obonet` node's data."""
    for prop_value_type in data.get("property_value", []):
        try:
            prop, value_type = prop_value_type.split(" ", 1)
        except ValueError:
            logger.info("malformed property: %s on %s", prop_value_type, term and term.curie)
            continue
        if property_prefix is not None and prop.startswith(property_prefix):
            prop = prop[len(property_prefix) :]

        try:
            value, _ = value_type.rsplit(" ", 1)  # second entry is the value type
        except ValueError:
            # logger.debug(f'property missing datatype. defaulting to string - {prop_value_type}')
            value = value_type  # could assign type to be 'xsd:string' by default
        value = value.strip('"')
        yield prop, value


def iterate_node_parents(
    data: Mapping[str, Any],
    *,
    prefix: str,
    identifier: str,
    strict: bool = True,
) -> Iterable[Reference]:
    """Extract parents from a :mod:`obonet` node's data."""
    for parent_curie in data.get("is_a", []):
        reference = Reference.from_curie(parent_curie, strict=strict)
        if reference is None:
            logger.warning(
                "[%s:%s] could not parse parent curie: %s", prefix, identifier, parent_curie
            )
            continue
        yield reference


def iterate_node_alt_ids(data: Mapping[str, Any], *, strict: bool = True) -> Iterable[Reference]:
    """Extract alternate identifiers from a :mod:`obonet` node's data."""
    for curie in data.get("alt_id", []):
        reference = Reference.from_curie(curie, strict=strict)
        if reference is not None:
            yield reference


def iterate_node_relationships(
    data: Mapping[str, Any],
    *,
    prefix: str,
    identifier: str,
    strict: bool = True,
) -> Iterable[Tuple[Reference, Reference]]:
    """Extract relationships from a :mod:`obonet` node's data."""
    for s in data.get("relationship", []):
        relation_curie, target_curie = s.split(" ")
        relation_prefix: Optional[str]
        relation_identifier: Optional[str]
        if relation_curie in RELATION_REMAPPINGS:
            relation_prefix, relation_identifier = RELATION_REMAPPINGS[relation_curie]
        else:
            relation_prefix, relation_identifier = normalize_curie(relation_curie)
        if relation_prefix is not None and relation_identifier is not None:
            relation = Reference(prefix=relation_prefix, identifier=relation_identifier)
        elif prefix is not None:
            relation = Reference(prefix=prefix, identifier=relation_curie)
        else:
            logger.debug("unhandled relation: %s", relation_curie)
            relation = Reference(prefix="obo", identifier=relation_curie)

        # TODO replace with omni-parser from :mod:`curies`
        target = Reference.from_curie(target_curie, strict=strict)
        if target is None:
            logger.warning(
                "[%s:%s] %s could not parse target %s", prefix, identifier, relation, target_curie
            )
            continue

        yield relation, target


def iterate_node_xrefs(
    *, prefix: str, data: Mapping[str, Any], strict: bool = True
) -> Iterable[Reference]:
    """Extract xrefs from a :mod:`obonet` node's data."""
    for xref in data.get("xref", []):
        xref = xref.strip()

        if curie_has_blacklisted_prefix(xref) or curie_is_blacklisted(xref) or ":" not in xref:
            continue  # sometimes xref to self... weird

        xref = remap_prefix(xref)

        split_space = " " in xref
        if split_space:
            _xref_split = xref.split(" ", 1)
            if _xref_split[1][0] not in {'"', "("}:
                logger.debug("[%s] Problem with space in xref %s", prefix, xref)
                continue
            xref = _xref_split[0]

        yv = Reference.from_curie(xref, strict=strict)
        if yv is not None:
            yield yv



# -*- coding: utf-8 -*-

"""Test iteration tools."""

import unittest

from pyobo.identifier_utils import normalize_curie
from pyobo.utils.iter import iterate_together


class TestIdentifierUtils(unittest.TestCase):
    """Test identifier utilities."""

    def test_strip_prefix(self):
        """Test stripping prefixes works."""
        self.assertEqual(("go", "1234"), normalize_curie("GO:1234"))
        self.assertEqual(("go", "1234"), normalize_curie("go:1234"))

        self.assertEqual((None, None), normalize_curie("1234"))
        self.assertEqual(("go", "1234"), normalize_curie("GO:GO:1234"))

        self.assertEqual(("pubmed", "1234"), normalize_curie("pubmed:1234"))
        # Test remapping
        self.assertEqual(("pubmed", "1234"), normalize_curie("pmid:1234"))
        self.assertEqual(("pubmed", "1234"), normalize_curie("PMID:1234"))


class TestIterate(unittest.TestCase):
    """Test iteration tools."""

    def test_a(self):
        """Test iterating two iterables together."""
        a = iter(
            [
                ("1", "a"),
                ("2", "b"),
                ("3", "c"),
            ]
        )
        b = iter(
            [
                ("1", "a1"),
                ("1", "a2"),
                ("2", "b1"),
                ("3", "c1"),
                ("3", "c2"),
            ]
        )
        rv = [
            ("1", "a", ["a1", "a2"]),
            ("2", "b", ["b1"]),
            ("3", "c", ["c1", "c2"]),
        ]

        r = iterate_together(a, b)
        self.assertNotIsInstance(r, list)
        self.assertEqual(rv, list(r))



# -*- coding: utf-8 -*-

"""Sources of OBO content."""

from class_resolver import ClassResolver

from .antibodyregistry import AntibodyRegistryGetter
from .ccle import CCLEGetter
from .cgnc import CGNCGetter
from .chembl import ChEMBLCompoundGetter
from .complexportal import ComplexPortalGetter
from .conso import CONSOGetter
from .cpt import CPTGetter
from .cvx import CVXGetter
from .depmap import DepMapGetter
from .dictybase_gene import DictybaseGetter
from .drugbank import DrugBankGetter
from .drugbank_salt import DrugBankSaltGetter
from .drugcentral import DrugCentralGetter
from .expasy import ExpasyGetter
from .famplex import FamPlexGetter
from .flybase import FlyBaseGetter
from .gwascentral_phenotype import GWASCentralPhenotypeGetter
from .gwascentral_study import GWASCentralStudyGetter
from .hgnc import HGNCGetter
from .hgncgenefamily import HGNCGroupGetter
from .icd10 import ICD10Getter
from .icd11 import ICD11Getter
from .interpro import InterProGetter
from .itis import ITISGetter
from .kegg import KEGGGeneGetter, KEGGGenomeGetter, KEGGPathwayGetter
from .mesh import MeSHGetter
from .mgi import MGIGetter
from .mirbase import MiRBaseGetter
from .mirbase_family import MiRBaseFamilyGetter
from .mirbase_mature import MiRBaseMatureGetter
from .msigdb import MSigDBGetter
from .ncbigene import NCBIGeneGetter
from .npass import NPASSGetter
from .pathbank import PathBankGetter
from .pfam import PfamGetter
from .pfam_clan import PfamClanGetter
from .pid import PIDGetter
from .pombase import PomBaseGetter
from .pubchem import PubChemCompoundGetter
from .reactome import ReactomeGetter
from .rgd import RGDGetter
from .rhea import RheaGetter
from .selventa import SCHEMGetter, SCOMPGetter, SDISGetter, SFAMGetter
from .sgd import SGDGetter
from .slm import SLMGetter
from .umls import UMLSGetter
from .uniprot import UniProtGetter, UniProtPtmGetter
from .wikipathways import WikiPathwaysGetter
from .zfin import ZFINGetter
from ..struct import Obo

__all__ = [
    "AntibodyRegistryGetter",
    "CCLEGetter",
    "CGNCGetter",
    "CONSOGetter",
    "CPTGetter",
    "CVXGetter",
    "ChEMBLCompoundGetter",
    "ComplexPortalGetter",
    "DepMapGetter",
    "DictybaseGetter",
    "DrugBankGetter",
    "DrugBankSaltGetter",
    "DrugCentralGetter",
    "ExpasyGetter",
    "FamPlexGetter",
    "FlyBaseGetter",
    "GWASCentralPhenotypeGetter",
    "GWASCentralStudyGetter",
    "HGNCGetter",
    "HGNCGroupGetter",
    "ICD10Getter",
    "ICD11Getter",
    "ITISGetter",
    "InterProGetter",
    "KEGGGeneGetter",
    "KEGGGenomeGetter",
    "KEGGPathwayGetter",
    "MGIGetter",
    "MSigDBGetter",
    "MeSHGetter",
    "MiRBaseFamilyGetter",
    "MiRBaseGetter",
    "MiRBaseMatureGetter",
    "NCBIGeneGetter",
    "NPASSGetter",
    "PIDGetter",
    "PathBankGetter",
    "PfamClanGetter",
    "PfamGetter",
    "PomBaseGetter",
    "PubChemCompoundGetter",
    "RGDGetter",
    "ReactomeGetter",
    "RheaGetter",
    "SCHEMGetter",
    "SCOMPGetter",
    "SDISGetter",
    "SFAMGetter",
    "SGDGetter",
    "SLMGetter",
    "UMLSGetter",
    "UniProtGetter",
    "UniProtPtmGetter",
    "WikiPathwaysGetter",
    "ZFINGetter",
    "ontology_resolver",
]


def _assert_sorted():
    _sorted = sorted(__all__)
    if _sorted != __all__:
        raise ValueError(f"unsorted. should be:\n{_sorted}")


_assert_sorted()
del _assert_sorted

ontology_resolver: ClassResolver[Obo] = ClassResolver.from_subclasses(base=Obo, suffix="Getter")
for getter in list(ontology_resolver):
    ontology_resolver.synonyms[getter.ontology] = getter



# -*- coding: utf-8 -*-

"""Tests for the OBO data structures."""

import unittest

from pyobo import Obo, Reference
from pyobo.struct.struct import BioregistryError


class Nope(Obo):
    """A class that will fail."""

    ontology = "nope"

    def iter_terms(self, force: bool = False):
        """Do not do anything."""


class TestStruct(unittest.TestCase):
    """Tests for the OBO data structures."""

    def test_invalid_prefix(self):
        """Test raising an error when an invalid prefix is used."""
        with self.assertRaises(BioregistryError):
            Nope()

    def test_reference_validation(self):
        """Test validation of prefix."""
        with self.assertRaises(ValueError):
            Reference(prefix="nope", identifier="also_nope")



# -*- coding: utf-8 -*-

"""Load the manually curated metaregistry."""

import itertools as itt
import json
import os
from functools import lru_cache
from pathlib import Path
from typing import Iterable, Mapping, Set, Tuple

import bioregistry

from ..constants import GLOBAL_SKIP, RAW_DIRECTORY

HERE = Path(__file__).parent.resolve()
CURATED_REGISTRY_PATH = HERE.joinpath("metaregistry.json")
CURATED_REGISTRY = json.loads(CURATED_REGISTRY_PATH.read_text())


def has_no_download(prefix: str) -> bool:
    """Return if the prefix is not available."""
    prefix_norm = bioregistry.normalize_prefix(prefix)
    return prefix_norm is not None and prefix_norm in _no_download()


@lru_cache(maxsize=1)
def _no_download() -> Set[str]:
    """Get the list of prefixes not available as OBO."""
    return {
        prefix
        for prefix in bioregistry.read_registry()
        if bioregistry.get_obo_download(prefix) is None
        and bioregistry.get_owl_download(prefix) is None
    }


def curie_has_blacklisted_prefix(curie: str) -> bool:
    """Check if the CURIE string has a blacklisted prefix."""
    return any(curie.startswith(x) for x in get_xrefs_prefix_blacklist())


@lru_cache(maxsize=1)
def get_xrefs_prefix_blacklist() -> Set[str]:
    """Get the set of blacklisted xref prefixes."""
    #: Xrefs starting with these prefixes will be ignored
    prefixes = set(
        itt.chain.from_iterable(CURATED_REGISTRY["blacklists"]["resource_prefix"].values())
    ) | set(CURATED_REGISTRY["blacklists"]["prefix"])
    nonsense = {
        prefix
        for prefix in prefixes
        if bioregistry.normalize_prefix(prefix.rstrip(":")) is not None
    }
    if nonsense:
        raise ValueError(
            f"The following prefixes were blacklisted but are in the bioregistry: {nonsense}"
        )
    return prefixes


def curie_has_blacklisted_suffix(curie: str) -> bool:
    """Check if the CURIE string has a blacklisted suffix."""
    return any(curie.endswith(suffix) for suffix in get_xrefs_suffix_blacklist())


@lru_cache(maxsize=1)
def get_xrefs_suffix_blacklist() -> Set[str]:
    """Get the set of blacklisted xref suffixes."""
    #: Xrefs ending with these suffixes will be ignored
    return set(CURATED_REGISTRY["blacklists"]["suffix"])


def curie_is_blacklisted(curie: str) -> bool:
    """Check if the full CURIE string is blacklisted."""
    return curie in get_xrefs_blacklist()


@lru_cache(maxsize=1)
def get_xrefs_blacklist() -> Set[str]:
    """Get the set of blacklisted xrefs."""
    rv = set()
    for x in CURATED_REGISTRY["blacklists"]["full"]:
        if isinstance(x, str):
            rv.add(x)
        elif isinstance(x, dict):
            if x.get("type") == "group":
                rv.update(x["text"])
            elif "text" in x:
                rv.add(x["text"])
            else:
                raise ValueError("invalid schema")
        else:
            raise TypeError
    return rv


@lru_cache(maxsize=1)
def get_remappings_full() -> Mapping[str, str]:
    """Get the remappings for xrefs based on the entire xref database."""
    return CURATED_REGISTRY["remappings"]["full"]


def remap_full(x: str) -> str:
    """Remap the string if possible otherwise return it."""
    return get_remappings_full().get(x, x)


@lru_cache(maxsize=1)
def get_remappings_prefix() -> Mapping[str, str]:
    """Get the remappings for xrefs based on the prefix.

    .. note:: Doesn't take into account the semicolon `:`
    """
    return CURATED_REGISTRY["remappings"]["prefix"]


def remap_prefix(curie: str) -> str:
    """Remap a prefix."""
    for old_prefix, new_prefix in get_remappings_prefix().items():
        if curie.startswith(old_prefix):
            return new_prefix + curie[len(old_prefix) :]
    return curie


def iter_cached_obo() -> Iterable[Tuple[str, str]]:
    """Iterate over cached OBO paths."""
    for prefix in os.listdir(RAW_DIRECTORY):
        if prefix in GLOBAL_SKIP or has_no_download(prefix) or bioregistry.is_deprecated(prefix):
            continue
        d = RAW_DIRECTORY.joinpath(prefix)
        if not os.path.isdir(d):
            continue
        for x in os.listdir(d):
            if x.endswith(".obo"):
                p = os.path.join(d, x)
                yield prefix, p



# -*- coding: utf-8 -*-

"""Converter for NCI PID."""

import logging
from collections import defaultdict
from typing import Iterable, List, Mapping, Tuple

import pandas as pd
from protmapper.uniprot_client import get_gene_name, get_hgnc_id

from ..api import get_id_name_mapping
from ..struct import Obo, Reference, Term
from ..struct.typedef import has_participant
from ..utils.ndex_utils import CX, ensure_ndex_network_set, iterate_aspect

__all__ = [
    "PIDGetter",
]

logger = logging.getLogger(__name__)

PREFIX = "pid.pathway"
NDEX_NETWORK_SET_UUID = "8a2d7ee9-1513-11e9-bb6a-0ac135e8bacf"

# Unused, but maybe useful for later
URL = (
    "https://github.com/NCIP/pathway-interaction-database/raw/master/download/NCI-Pathway-Info.xlsx"
)


class PIDGetter(Obo):
    """An ontology representation of the NCI's Pathway Interaction Database."""

    ontology = PREFIX
    static_version = "1.0.0"
    typedefs = [has_participant]

    def iter_terms(self, force: bool = False) -> Iterable[Term]:
        """Iterate over terms in the ontology."""
        return iter_terms()


def get_obo() -> Obo:
    """Get NCI PID as OBO."""
    return PIDGetter()


def iter_networks(use_tqdm: bool = False, force: bool = False) -> Iterable[Tuple[str, CX]]:
    """Iterate over NCI PID networks."""
    yield from ensure_ndex_network_set(
        PREFIX, NDEX_NETWORK_SET_UUID, use_tqdm=use_tqdm, force=force
    )


def iter_terms(force: bool = False) -> Iterable[Term]:
    """Iterate over NCI PID terms."""
    hgnc_id_to_name = get_id_name_mapping("hgnc")
    hgnc_name_to_id = {v: k for k, v in hgnc_id_to_name.items()}

    for uuid, cx in iter_networks(force=force, use_tqdm=True):
        name = None
        for node in iterate_aspect(cx, "networkAttributes"):
            if node["n"] == "name":
                name = node["v"]

        term = Term(
            reference=Reference(prefix=PREFIX, identifier=uuid, name=name),
        )

        genes = set()
        for node in iterate_aspect(cx, "nodes"):
            name, reference = node["n"], node["r"]
            hgnc_id = hgnc_name_to_id.get(name)
            if hgnc_id:
                genes.add((hgnc_id, name))
            elif any(reference.startswith(x) for x in ("CHEBI:", "cas:")):
                pass
            elif reference.startswith("uniprot:"):
                uniprot_id = reference[len("uniprot:") :]
                hgnc_id = get_hgnc_id(uniprot_id)
                if hgnc_id is None:  # this only happens for proteins that seem to be virus related
                    # TODO reinvestigate this later
                    logger.debug(
                        "uniprot could not map %s/%s/%s to HGNC",
                        name,
                        reference,
                        get_gene_name(uniprot_id, web_fallback=False),
                    )
                else:
                    name = hgnc_id_to_name[hgnc_id]
                    genes.add((hgnc_id, name))
            else:
                logger.debug(f"unmapped: {name}, {reference}")

        for hgnc_id, hgnc_symbol in genes:
            term.append_relationship(
                has_participant, Reference(prefix="hgnc", identifier=hgnc_id, name=hgnc_symbol)
            )

        yield term


# TODO Update mappings based on curated sheet on google

MAPPINGS_SHEET = (
    "https://docs.google.com/spreadsheets/d/e/2PACX-1vR7CuRJYBEGMobZQpjmzixL"
    "2bfWe4l4FRTySCSdQ_vwvkhpH_9rL4w1IMAe7ZkvhUcfCtQubVmRxqdW/pub?output=tsv"
)


def get_curation_df() -> pd.DataFrame:
    """Get the curated dataframe."""
    df = pd.read_csv(MAPPINGS_SHEET, sep="\t")
    df = df[df["Namespace"].notna()]
    return df[["Text from NDEx", "Type", "Namespace", "Identifier"]]


def get_remapping() -> Mapping[str, List[Tuple[str, str]]]:
    """Get a mapping from text to list of HGNC id/symbols."""
    curation_df = get_curation_df()
    rv = defaultdict(list)
    for text, dtype, prefix, identifier in curation_df.values:
        if dtype == "protein":
            if prefix == "hgnc":
                rv[text] = [identifier]
                continue
        elif dtype == "protein family":
            if prefix == "fplx":
                logger.debug("unhandled - famplex protein family")
            elif prefix == "hgnc.genefamily":
                logger.debug("unhandled - HGNC gene family")
            else:
                logger.debug("unhandled prefix - %s", prefix)
        else:
            logger.debug("unhandled type - %s", dtype)

    return rv


if __name__ == "__main__":
    PIDGetter.cli()



# -*- coding: utf-8 -*-

"""Tests for identifiers.org URL generation."""

import logging
import unittest

import requests
from bioregistry import get_identifiers_org_iri

logger = logging.getLogger(__name__)

#: These resources don't seem to exist anymore
BLACKLIST = {
    "abs",
    "aftol.taxonomy",
    "agricola",
    "ecogene",
    "euclinicaltrials",
    "fsnp",
    "gold",
    "gold.genome",
    "gold.meta",
}

#: These resources will need special rules for resolving
UNSOLVED = {
    "ark",
    "did",
    "gramene.growthstage",
    "gwascentral.phenotype",
    # TODO
}


class TestMiriam(unittest.TestCase):
    """Test generating identifiers.org links."""

    def test_successful(self):
        """Test CURIEs that should work."""
        curies = [
            ("go", "0006915"),  # name in LUI
            ("doid", "11337"),  # name in LUI
            ("mesh", "C000100"),  # namespace not in LUI
        ]

        # curies = []
        # for entry in get_miriam():
        #     prefix = entry['prefix']
        #     if prefix <= 'gramene.growthstage':  # TODO REMOVE THIS LINE
        #         continue  # TODO REMOVE THIS LINE
        #     norm_prefix = normalize_prefix(prefix)
        #     self.assertIsNotNone(norm_prefix, msg=f'could not normalize MIRIAM prefix: {norm_prefix}')
        #     curies.append((prefix, norm_prefix, entry['sampleId']))

        for prefix, identifier in curies:
            if prefix in BLACKLIST or prefix in UNSOLVED:
                continue
            with self.subTest(prefix=prefix, msg=f"failed for MIRIAM prefix: {prefix}"):
                url = get_identifiers_org_iri(prefix, identifier)
                self.assertIsNotNone(url, msg=f"metaregistry does not contain prefix {prefix}")
                try:
                    res = requests.get(url)
                except (
                    requests.exceptions.SSLError,
                    requests.exceptions.ConnectionError,
                ):
                    logger.warning(f"identifiers.org has a problem resolving prefix {prefix}")
                    continue
                self.assertFalse(
                    res.text.startswith("INVALID"),
                    msg=f"invalid url for {prefix}: {url}\n\n{res.text}",
                )

    def test_unsuccessful(self):
        """Test links that should fail."""
        curies = [
            ("nope_nope_nope", "0006915"),
        ]
        for prefix, identifier in curies:
            with self.subTest(prefix=prefix):
                url = get_identifiers_org_iri(prefix, identifier)
                self.assertIsNone(url)



# -*- coding: utf-8 -*-

"""Tests for PyOBO caches."""

import os
import time
import unittest
from tempfile import TemporaryDirectory

from pyobo.utils.cache import cached_mapping, cached_multidict
from pyobo.utils.io import open_map_tsv, open_multimap_tsv

sleep_time = 3


class TestCaches(unittest.TestCase):
    """Tests for PyOBO cache decorators."""

    def test_mapping(self):
        """Test the mapping cache."""
        with TemporaryDirectory() as directory:
            path = os.path.join(directory, "test.tsv")
            header = ["key", "value"]

            @cached_mapping(path=path, header=header)
            def _get_mapping():
                time.sleep(sleep_time)
                return dict(a="x", b="y", c="z")

            start_time = time.time()
            rv1 = _get_mapping()
            elapsed = time.time() - start_time
            self.assertGreater(elapsed, sleep_time)

            self._help_test_mapping(rv1)

            """Test cache"""
            rv3 = open_map_tsv(path)
            self._help_test_mapping(rv3)

            """Test reload"""

            start_time = time.time()
            rv2 = _get_mapping()  # this time should be fast
            elapsed = time.time() - start_time
            self.assertLess(elapsed, sleep_time)

            self._help_test_mapping(rv2)

    def _help_test_mapping(self, d):
        self.assertIsNotNone(d)
        self.assertEqual(3, len(d))
        self.assertEqual(dict(a="x", b="y", c="z"), d)

    def test_multidict(self):
        """Test caching a multidict."""
        with TemporaryDirectory() as directory:
            path = os.path.join(directory, "test.tsv")
            header = ["key", "value"]

            @cached_multidict(path=path, header=header)
            def _get_multidict():
                time.sleep(sleep_time)
                return dict(a=["a1", "a2"], b=["b1"], c=["c1", "c2"])

            start_time = time.time()
            rv1 = _get_multidict()
            elapsed = time.time() - start_time
            self.assertGreater(elapsed, sleep_time)

            self._help_test_multidict(rv1)

            """Test cache"""
            rv3 = open_multimap_tsv(path)
            self._help_test_multidict(rv3)

            """Test reload"""
            start_time = time.time()
            rv2 = _get_multidict()  # this time should be fast
            elapsed = time.time() - start_time
            self.assertLess(elapsed, sleep_time)

            self._help_test_multidict(rv2)

    def _help_test_multidict(self, d):
        self.assertIsNotNone(d)
        self.assertEqual(3, len(d))
        self.assertEqual(dict(a=["a1", "a2"], b=["b1"], c=["c1", "c2"]), d)



# -*- coding: utf-8 -*-

"""Loading of the relations ontology names."""

import csv
import os
from functools import lru_cache
from typing import Mapping, Tuple

import requests

__all__ = [
    "load_ro",
]

HERE = os.path.abspath(os.path.dirname(__file__))
PATH = os.path.join(HERE, "ro.tsv")
URL = "http://purl.obolibrary.org/obo/ro.json"
PREFIX = "http://purl.obolibrary.org/obo/"


@lru_cache(maxsize=1)
def load_ro() -> Mapping[Tuple[str, str], str]:
    """Load the relation ontology names."""
    if not os.path.exists(PATH):
        download()
    with open(PATH) as file:
        return {
            (prefix, identifier): name
            for prefix, identifier, name in csv.reader(file, delimiter="\t")
        }


def download():
    """Download the latest version of the Relation Ontology."""
    rows = []
    res_json = requests.get(URL).json()
    for node in res_json["graphs"][0]["nodes"]:
        identifier = node["id"]
        if not identifier.startswith(PREFIX):
            continue
        identifier = identifier[len(PREFIX) :]
        if all(not identifier.startswith(p) for p in ("RO", "BFO", "UPHENO")):
            continue
        prefix, identifier = identifier.split("_", 1)
        name = node.get("lbl")
        if name:
            rows.append((prefix.lower(), identifier, name))

    with open(PATH, "w") as file:
        writer = csv.writer(file, delimiter="\t")
        writer.writerows(sorted(rows))


if __name__ == "__main__":
    download()



"""Static site generator."""

import itertools as itt
from collections import defaultdict
from operator import attrgetter
from pathlib import Path
from typing import Optional, Sequence, Tuple, Union

import bioregistry
from bioregistry.constants import BIOREGISTRY_DEFAULT_BASE_URL
from jinja2 import Environment, FileSystemLoader
from tqdm.auto import tqdm

from pyobo import Obo
from pyobo.struct import part_of

__all__ = [
    "make_site",
]

HERE = Path(__file__).parent.resolve()
environment = Environment(
    autoescape=True, loader=FileSystemLoader(HERE), trim_blocks=True, lstrip_blocks=True
)
environment.globals["bioregistry"] = bioregistry
term_template = environment.get_template("term.html")
typedef_template = environment.get_template("typedef.html")
index_template = environment.get_template("index.html")


def make_site(
    obo: Obo,
    directory: Union[str, Path],
    use_subdirectories: bool = True,
    manifest: bool = False,
    resource: Optional[bioregistry.Resource] = None,
    metaregistry_metaprefix: Optional[str] = None,
    metaregistry_name: Optional[str] = None,
    metaregistry_base_url: Optional[str] = None,
    show_properties_in_manifest: Optional[Sequence[Tuple[str, str]]] = None,
) -> None:
    """Make a website in the given directory.

    :param obo: The ontology to generate a site for
    :param directory: The directory in which to generate the site
    :param use_subdirectories: If true, creates directories for each term/property/typedef with an index.html
        inside. If false, creates HTML files named with the identifiers.
    :param manifest: If true, lists all entries on the homepage.
    :param resource: A custom resource
    """
    directory = Path(directory)
    directory.mkdir(exist_ok=True, parents=True)

    if metaregistry_metaprefix is None:
        metaregistry_metaprefix = "bioregistry"
    if metaregistry_name is None:
        metaregistry_name = "Bioregistry"
    if metaregistry_base_url is None:
        metaregistry_base_url = BIOREGISTRY_DEFAULT_BASE_URL
    metaregistry_base_url = metaregistry_base_url.rstrip("/")

    if resource is None:
        resource = bioregistry.get_resource(obo.ontology)
    if resource is None:
        raise KeyError

    if not manifest:
        _manifest = None
    else:
        _manifest = sorted(
            (term for term in itt.chain(obo, obo.typedefs or []) if term.prefix == obo.ontology),
            key=attrgetter("identifier"),
        )

    terms = [term for term in obo if term.prefix == obo.ontology]

    directory.joinpath("index.html").write_text(
        index_template.render(
            obo=obo,
            resource=resource,
            manifest=_manifest,
            number_terms=len(terms),
            manager=bioregistry.manager,
            metaregistry_metaprefix=metaregistry_metaprefix,
            metaregistry_name=metaregistry_name,
            metaregistry_base_url=metaregistry_base_url,
        )
    )

    parent_to_child = defaultdict(list)
    for term in tqdm(terms, desc=f"{obo.ontology} caching parents", unit="term", unit_scale=True):
        for parent in term.parents or []:
            parent_to_child[parent.curie].append(term)

    parts = defaultdict(list)
    for term in tqdm(terms, desc=f"{obo.ontology} caching parents", unit="term", unit_scale=True):
        for whole in term.get_relationships(part_of):
            parts[whole.curie].append(term)

    for term in tqdm(terms, desc=f"{obo.ontology} website", unit="term", unit_scale=True):
        if use_subdirectories:
            subdirectory = directory.joinpath(term.identifier)
            subdirectory.mkdir(exist_ok=True, parents=True)
            path = subdirectory.joinpath("index.html")
        else:
            path = directory.joinpath(term.identifier).with_suffix(".html")
        path.write_text(
            term_template.render(
                term=term,
                obo=obo,
                resource=resource,
                children=parent_to_child.get(term.curie),
                parts=parts.get(term.curie),
                manager=bioregistry.manager,
                metaregistry_metaprefix=metaregistry_metaprefix,
                metaregistry_name=metaregistry_name,
                metaregistry_base_url=metaregistry_base_url,
            )
        )

    for typedef in obo.typedefs or []:
        if typedef.prefix != obo.ontology:
            continue
        if use_subdirectories:
            subdirectory = directory.joinpath(typedef.identifier)
            subdirectory.mkdir(exist_ok=True, parents=True)
            path = subdirectory.joinpath("index.html")
        else:
            path = directory.joinpath(typedef.identifier).with_suffix(".html")
        path.write_text(
            typedef_template.render(
                typedef=typedef,
                obo=obo,
                resource=resource,
                manager=bioregistry.manager,
                metaregistry_metaprefix=metaregistry_metaprefix,
                metaregistry_name=metaregistry_name,
                metaregistry_base_url=metaregistry_base_url,
            )
        )


def _main():
    from pyobo.sources import HGNCGetter, UniProtPtmGetter

    for cls in [UniProtPtmGetter, HGNCGetter]:
        obo = cls()
        make_site(obo, f"/Users/cthoyt/Desktop/pyobo-sites/{obo.ontology}/")


if __name__ == "__main__":
    _main()



# -*- coding: utf-8 -*-

"""Utilities for high-level API."""

import json
from typing import Optional

import bioversions

from ..utils.path import prefix_directory_join

__all__ = [
    "get_version",
]


def get_version(prefix: str) -> Optional[str]:
    """Get the version for the resource, if available.

    :param prefix: the resource name
    :return: The version if available else None
    """
    try:
        version = bioversions.get_version(prefix)
    except IOError:
        raise IOError(f"[{prefix}] could not get version from bioversions")
    except KeyError:
        pass  # this prefix isn't available from bioversions
    else:
        if version:
            return version

    metadata_json_path = prefix_directory_join(prefix, name="metadata.json", ensure_exists=False)
    if metadata_json_path.exists():
        data = json.loads(metadata_json_path.read_text())
        return data["version"]

    return None



# -*- coding: utf-8 -*-

"""Use synonyms from OBO to normalize names."""

import logging
from abc import ABC, abstractmethod
from dataclasses import dataclass
from functools import lru_cache
from typing import Dict, Iterable, List, Mapping, Optional, Set, Tuple, Union

import bioregistry

from .api import names
from .utils.io import multisetdict

__all__ = [
    "ground",
    "Normalizer",
    "OboNormalizer",
    "MultiNormalizer",
    "NormalizationResult",
]

logger = logging.getLogger(__name__)

NormalizationSuccess = Tuple[str, str, str]
NormalizationFailure = Tuple[None, None, str]
NormalizationResult = Union[NormalizationSuccess, NormalizationFailure]


class Normalizer(ABC):
    """A normalizer."""

    id_to_name: Dict[str, str]
    id_to_synonyms: Dict[str, List[str]]

    #: A mapping from all synonyms to the set of identifiers that they point to.
    #: In a perfect world, each would only be a single element.
    synonym_to_identifiers_mapping: Dict[str, Set[str]]
    #: A mapping from normalized names to the actual ones that they came from
    norm_name_to_name: Dict[str, Set[str]]

    def __init__(
        self,
        id_to_name: Dict[str, str],
        id_to_synonyms: Dict[str, List[str]],
        remove_prefix: Optional[str] = None,
    ) -> None:  # noqa: D107
        """Initialize the normalizer.

        :param id_to_name: An identifier to name dictionary.
        :param id_to_synonyms: An identifier to list of synonyms dictionary.
        :param remove_prefix: A prefix to be removed from the identifiers. Useful for nomenclatures like ChEBI.
        """
        self.id_to_name = id_to_name
        self.id_to_synonyms = id_to_synonyms
        self.synonym_to_identifiers_mapping = multisetdict(
            self._iterate_synonyms_to_identifiers(
                id_to_name=self.id_to_name,
                id_to_synonyms=self.id_to_synonyms,
                remove_prefix=remove_prefix,
            )
        )
        self.norm_name_to_name = self._get_norm_name_to_names(self.synonym_to_identifiers_mapping)

    @classmethod
    def _get_norm_name_to_names(cls, synonyms: Iterable[str]) -> Dict[str, Set[str]]:
        return multisetdict((cls._normalize_text(synonym), synonym) for synonym in synonyms)

    @staticmethod
    def _normalize_text(text: str) -> str:
        text = text.strip().strip('"').strip("'").lower()
        text = normalize_dashes(text)
        text = text.replace("-", "")  # remove all dashes
        text = text.replace(" ", "")  # remove all spaces
        return text

    @staticmethod
    def _iterate_synonyms_to_identifiers(
        *,
        id_to_name: Mapping[str, str],
        id_to_synonyms: Mapping[str, Iterable[str]],
        remove_prefix: Optional[str] = None,
    ) -> Iterable[Tuple[str, str]]:
        if remove_prefix is not None:
            remove_prefix = f'{remove_prefix.lower().rstrip(":")}:'

        # Add name
        for identifier, name in id_to_name.items():
            if remove_prefix and identifier.lower().startswith(remove_prefix):
                identifier = identifier[len(remove_prefix) :]

            yield name, identifier

        # Add synonyms
        for identifier, synonyms in id_to_synonyms.items():
            if remove_prefix and identifier.lower().startswith(remove_prefix):
                identifier = identifier[len(remove_prefix) :]

            for synonym in synonyms:
                # it might overwrite but this is probably always due to alternate ids
                yield synonym, identifier

    def get_names(self, query: str) -> List[str]:
        """Get all names to which the query text maps."""
        norm_text = self._normalize_text(query)
        return list(self.norm_name_to_name.get(norm_text, []))

    @abstractmethod
    def normalize(self, query: str) -> NormalizationResult:
        """Try and normalize a name to a identifier and canonical name."""
        raise NotImplementedError


@lru_cache()
def get_normalizer(prefix: str) -> Normalizer:
    """Get an OBO normalizer."""
    norm_prefix = bioregistry.normalize_prefix(prefix)
    if norm_prefix is None:
        raise ValueError(f"unhandled prefix: {prefix}")
    logger.info("getting obo normalizer for %s", norm_prefix)
    normalizer = OboNormalizer(norm_prefix)
    logger.debug(
        "normalizer for %s with %s name lookups",
        normalizer.prefix,
        len(normalizer.norm_name_to_name),
    )
    return normalizer


def ground(prefix: Union[str, Iterable[str]], query: str) -> NormalizationResult:
    """Normalize a string given the prefix's labels and synonyms.

    :param prefix: If a string, only grounds against that namespace. If a list, will try grounding
     against all in that order
    :param query: The string to try grounding
    """
    if isinstance(prefix, str):
        normalizer = get_normalizer(prefix)
        return normalizer.normalize(query)
    else:
        for p in prefix:
            norm_prefix, identifier, name = ground(p, query)
            if norm_prefix and identifier and name:
                return norm_prefix, identifier, name
        return None, None, query


class OboNormalizer(Normalizer):
    """A utility for normalizing by names."""

    def __init__(self, prefix: str) -> None:  # noqa: D107
        self.prefix = prefix
        self._len_prefix = len(prefix)
        id_to_name = names.get_id_name_mapping(prefix)
        id_to_synonyms = names.get_id_synonyms_mapping(prefix)
        super().__init__(
            id_to_name=dict(id_to_name),
            id_to_synonyms=dict(id_to_synonyms),
            remove_prefix=prefix,
        )

    def __repr__(self) -> str:  # noqa: D105
        return f'OboNormalizer(prefix="{self.prefix}")'

    def normalize(self, query: str) -> NormalizationResult:
        """Try and normalize a name to a identifier and canonical name."""
        names = self.get_names(query)
        if not names:
            return None, None, query

        for name in names:
            identifiers = self.synonym_to_identifiers_mapping[name]
            for identifier in identifiers:
                if identifier in self.id_to_name:
                    return self.prefix, identifier, self.id_to_name[identifier]
            logger.warning(f"Could not find valid identifier for {name} from {identifiers}")

        # maybe it happens that one can't be found?
        logger.warning(f"was able to look up name {query}->{names} but not find fresh identifier")
        return None, None, query


@dataclass
class MultiNormalizer:
    """Multiple normalizers together.

    If you're looking for taxa of exotic plants, you might use:

    >>> from pyobo.normalizer import MultiNormalizer
    >>> normalizer = MultiNormalizer(prefixes=['ncbitaxon', 'itis'])
    >>> normalizer.normalize('Homo sapiens')
    ('ncbitaxon', '9606', 'Homo sapiens')
    >>> normalizer.normalize('Abies bifolia')  # variety not listed in NCBI
    ('itis', '507501', 'Abies bifolia')
    >>> normalizer.normalize('vulcan')  # nice try, nerds
    (None, None, None)
    """

    #: The normalizers for each prefix
    normalizers: List[Normalizer]

    @staticmethod
    def from_prefixes(prefixes: List[str]) -> "MultiNormalizer":
        """Instantiate normalizers based on the given prefixes, in preferred order.."""
        return MultiNormalizer([get_normalizer(prefix) for prefix in prefixes])

    def normalize(self, query: str) -> NormalizationResult:
        """Try and normalize a canonical name using multiple normalizers."""
        for normalizer in self.normalizers:
            prefix, identifier, name = normalizer.normalize(query)
            if prefix and identifier and name:  # all not empty
                return prefix, identifier, name
        return None, None, query


# See: https://en.wikipedia.org/wiki/Dash
FIGURE_DASH = b"\xe2\x80\x92".decode("utf-8")
EN_DASH = b"\xe2\x80\x93".decode("utf-8")
EM_DASH = b"\xe2\x80\x94".decode("utf-8")
HORIZONAL_BAR = b"\xe2\x80\x95".decode("utf-8")
NORMAL_DASH = "-"


def normalize_dashes(s: str) -> str:
    """Normalize dashes in a string."""
    return (
        s.replace(FIGURE_DASH, NORMAL_DASH)
        .replace(EN_DASH, NORMAL_DASH)
        .replace(EM_DASH, NORMAL_DASH)
        .replace(HORIZONAL_BAR, NORMAL_DASH)
    )



# -*- coding: utf-8 -*-

"""Data structures for OBO."""

import gzip
import json
import logging
import os
from collections import defaultdict
from dataclasses import dataclass, field
from datetime import datetime
from operator import attrgetter
from pathlib import Path
from textwrap import dedent
from typing import (
    Any,
    Callable,
    ClassVar,
    Collection,
    Dict,
    Iterable,
    Iterator,
    List,
    Mapping,
    Optional,
    Sequence,
    Set,
    TextIO,
    Tuple,
    Union,
)

import bioregistry
import click
import networkx as nx
import pandas as pd
from more_click import force_option, verbose_option
from tqdm.auto import tqdm
from typing_extensions import Literal

from .reference import Reference, Referenced
from .typedef import (
    RelationHint,
    TypeDef,
    comment,
    default_typedefs,
    exact_match,
    from_species,
    get_reference_tuple,
    has_part,
    is_a,
    orthologous,
    part_of,
    see_also,
)
from .utils import comma_separate, obo_escape_slim
from ..constants import (
    DATE_FORMAT,
    NCBITAXON_PREFIX,
    RELATION_ID,
    RELATION_PREFIX,
    TARGET_ID,
    TARGET_PREFIX,
)
from ..identifier_utils import normalize_curie
from ..utils.io import multidict, write_iterable_tsv
from ..utils.misc import obo_to_owl
from ..utils.path import get_prefix_obo_path, prefix_directory_join

__all__ = [
    "Synonym",
    "SynonymTypeDef",
    "SynonymSpecificity",
    "SynonymSpecificities",
    "Term",
    "Obo",
    "make_ad_hoc_ontology",
]

logger = logging.getLogger(__name__)

SynonymSpecificity = Literal["EXACT", "NARROW", "BROAD", "RELATED"]
SynonymSpecificities: Sequence[SynonymSpecificity] = ("EXACT", "NARROW", "BROAD", "RELATED")


@dataclass
class Synonym:
    """A synonym with optional specificity and references."""

    #: The string representing the synonym
    name: str

    #: The specificity of the synonym
    specificity: SynonymSpecificity = "EXACT"

    #: The type of synonym. Must be defined in OBO document!
    type: "SynonymTypeDef" = field(
        default_factory=lambda: DEFAULT_SYNONYM_TYPE  # type:ignore
    )

    #: References to articles where the synonym appears
    provenance: List[Reference] = field(default_factory=list)

    def to_obo(self) -> str:
        """Write this synonym as an OBO line to appear in a [Term] stanza."""
        return f"synonym: {self._fp()}"

    def _fp(self) -> str:
        x = f'"{self._escape(self.name)}" {self.specificity}'
        if self.type and self.type.pair != DEFAULT_SYNONYM_TYPE.pair:
            x = f"{x} {self.type.curie}"
        return f"{x} [{comma_separate(self.provenance)}]"

    @staticmethod
    def _escape(s: str) -> str:
        return s.replace('"', '\\"')


@dataclass
class SynonymTypeDef(Referenced):
    """A type definition for synonyms in OBO."""

    reference: Reference
    specificity: Optional[SynonymSpecificity] = None

    def to_obo(self) -> str:
        """Serialize to OBO."""
        rv = f'synonymtypedef: {self.curie} "{self.name}"'
        if self.specificity:
            rv = f"{rv} {self.specificity}"
        return rv

    @classmethod
    def from_text(
        cls,
        text: str,
        specificity: Optional[SynonymSpecificity] = None,
        *,
        lower: bool = True,
    ) -> "SynonymTypeDef":
        """Get a type definition from text that's normalized."""
        identifier = (
            text.replace("-", "_")
            .replace(" ", "_")
            .replace('"', "")
            .replace(")", "")
            .replace("(", "")
        )
        if lower:
            identifier = identifier.lower()
        return cls(
            reference=Reference(prefix="obo", identifier=identifier, name=text.replace('"', "")),
            specificity=specificity,
        )


DEFAULT_SYNONYM_TYPE = SynonymTypeDef(
    reference=Reference(prefix="oboInOwl", identifier="SynonymType", name="Synonym"),
)

ReferenceHint = Union[Reference, "Term", Tuple[str, str], str]


def _ensure_ref(reference: ReferenceHint) -> Reference:
    if reference is None:
        raise ValueError("can not append null reference")
    if isinstance(reference, Term):
        return reference.reference
    if isinstance(reference, str):
        _rv = Reference.from_curie(reference)
        if _rv is None:
            raise ValueError
        return _rv
    if isinstance(reference, tuple):
        return Reference(*reference)
    if isinstance(reference, Reference):
        return reference
    raise TypeError


@dataclass
class Term(Referenced):
    """A term in OBO."""

    #: The primary reference for the entity
    reference: Reference

    #: A description of the entity
    definition: Optional[str] = None

    #: References to articles in which the term appears
    provenance: List[Reference] = field(default_factory=list)

    #: Relationships defined by [Typedef] stanzas
    relationships: Dict[TypeDef, List[Reference]] = field(default_factory=lambda: defaultdict(list))

    #: Properties, which are not defined with Typedef and have scalar values instead of references.
    properties: Dict[str, List[str]] = field(default_factory=lambda: defaultdict(list))

    #: Relationships with the default "is_a"
    parents: List[Reference] = field(default_factory=list)

    #: Synonyms of this term
    synonyms: List[Synonym] = field(default_factory=list)

    #: Equivalent references
    xrefs: List[Reference] = field(default_factory=list)
    xref_types: List[Reference] = field(default_factory=list)

    #: Alternate Identifiers
    alt_ids: List[Reference] = field(default_factory=list)

    #: The sub-namespace within the ontology
    namespace: Optional[str] = None

    #: An annotation for obsolescence. By default, is None, but this means that it is not obsolete.
    is_obsolete: Optional[bool] = None

    def __hash__(self):  # noqa: D105
        return hash((self.__class__, self.prefix, self.identifier))

    @classmethod
    def from_triple(
        cls,
        prefix: str,
        identifier: str,
        name: Optional[str] = None,
        definition: Optional[str] = None,
    ) -> "Term":
        """Create a term from a reference."""
        return cls(
            reference=Reference(prefix=prefix, identifier=identifier, name=name),
            definition=definition,
        )

    @classmethod
    def auto(
        cls,
        prefix: str,
        identifier: str,
    ) -> "Term":
        """Create a term from a reference."""
        from ..api import get_definition

        return cls(
            reference=Reference.auto(prefix=prefix, identifier=identifier),
            definition=get_definition(prefix, identifier),
        )

    @classmethod
    def from_curie(cls, curie: str, name: Optional[str] = None) -> "Term":
        """Create a term directly from a CURIE and optional name."""
        prefix, identifier = normalize_curie(curie)
        if prefix is None or identifier is None:
            raise ValueError
        return cls.from_triple(prefix=prefix, identifier=identifier, name=name)

    def append_provenance(self, reference: ReferenceHint) -> None:
        """Add a provenance reference."""
        self.provenance.append(_ensure_ref(reference))

    def append_synonym(
        self, synonym: Union[str, Synonym], type: Optional[SynonymTypeDef] = None
    ) -> None:
        """Add a synonym."""
        if isinstance(synonym, str):
            synonym = Synonym(synonym, type=type or DEFAULT_SYNONYM_TYPE)
        self.synonyms.append(synonym)

    def append_alt(self, alt: Union[str, Reference]) -> None:
        """Add an alternative identifier."""
        if isinstance(alt, str):
            alt = Reference(prefix=self.prefix, identifier=alt)
        self.alt_ids.append(alt)

    def append_see_also(self, reference: ReferenceHint) -> "Term":
        """Add a see also relationship."""
        self.relationships[see_also].append(_ensure_ref(reference))
        return self

    def append_comment(self, value: str) -> "Term":
        """Add a comment relationship."""
        self.append_property(comment.curie, value)
        return self

    def append_parent(self, reference: ReferenceHint) -> "Term":
        """Add a parent to this entity."""
        reference = _ensure_ref(reference)
        if reference not in self.parents:
            self.parents.append(reference)
        return self

    def extend_parents(self, references: Collection[Reference]) -> None:
        """Add a collection of parents to this entity."""
        if any(x is None for x in references):
            raise ValueError("can not append a collection of parents containing a null parent")
        self.parents.extend(references)

    def get_properties(self, prop) -> List[str]:
        """Get properties from the given key."""
        return self.properties[prop]

    def get_property(self, prop) -> Optional[str]:
        """Get a single property of the given key."""
        r = self.get_properties(prop)
        if not r:
            return None
        if len(r) != 1:
            raise ValueError
        return r[0]

    def get_relationship(self, typedef: TypeDef) -> Optional[Reference]:
        """Get a single relationship of the given type."""
        r = self.get_relationships(typedef)
        if not r:
            return None
        if len(r) != 1:
            raise ValueError
        return r[0]

    def get_relationships(self, typedef: TypeDef) -> List[Reference]:
        """Get relationships from the given type."""
        return self.relationships[typedef]

    def append_exact_match(self, reference):
        """Append an exact match, also adding an xref."""
        reference = _ensure_ref(reference)
        self.append_relationship(exact_match, reference)
        self.append_xref(reference)
        return self

    def append_xref(self, reference: ReferenceHint) -> None:
        """Append an xref."""
        self.xrefs.append(_ensure_ref(reference))

    def append_relationship(self, typedef: TypeDef, reference: ReferenceHint) -> None:
        """Append a relationship."""
        self.relationships[typedef].append(_ensure_ref(reference))

    def set_species(self, identifier: str, name: Optional[str] = None):
        """Append the from_species relation."""
        if name is None:
            from pyobo.resources.ncbitaxon import get_ncbitaxon_name

            name = get_ncbitaxon_name(identifier)
        self.append_relationship(
            from_species, Reference(prefix=NCBITAXON_PREFIX, identifier=identifier, name=name)
        )

    def get_species(self, prefix: str = NCBITAXON_PREFIX) -> Optional[Reference]:
        """Get the species if it exists.

        :param prefix: The prefix to use in case the term has several species annotations.
        """
        for species in self.relationships.get(from_species, []):
            if species.prefix == prefix:
                return species
        return None

    def extend_relationship(self, typedef: TypeDef, references: Iterable[Reference]) -> None:
        """Append several relationships."""
        if any(x is None for x in references):
            raise ValueError("can not extend a collection that includes a null reference")
        self.relationships[typedef].extend(references)

    def append_property(self, prop: str, value: str) -> None:
        """Append a property."""
        self.properties[prop].append(value)

    def _definition_fp(self) -> str:
        assert self.definition is not None
        return f'"{obo_escape_slim(self.definition)}" [{comma_separate(self.provenance)}]'

    def iterate_relations(self) -> Iterable[Tuple[TypeDef, Reference]]:
        """Iterate over pairs of typedefs and targets."""
        for typedef, targets in self.relationships.items():
            for target in targets:
                yield typedef, target

    def iterate_properties(self) -> Iterable[Tuple[str, str]]:
        """Iterate over pairs of property and values."""
        for prop, values in self.properties.items():
            for value in values:
                yield prop, value

    def iterate_obo_lines(self) -> Iterable[str]:
        """Iterate over the lines to write in an OBO file."""
        yield "\n[Term]"
        yield f"id: {self.curie}"
        if self.name:
            yield f"name: {obo_escape_slim(self.name)}"
        if self.namespace and self.namespace != "?":
            namespace_normalized = (
                self.namespace.replace(" ", "_").replace("-", "_").replace("(", "").replace(")", "")
            )
            yield f"namespace: {namespace_normalized}"

        if self.definition:
            yield f"def: {self._definition_fp()}"

        for xref in sorted(self.xrefs, key=attrgetter("prefix", "identifier")):
            yield f"xref: {xref}"

        for parent in sorted(self.parents, key=attrgetter("prefix", "identifier")):
            yield f"is_a: {parent}"

        for typedef, references in sorted(self.relationships.items(), key=_sort_relations):
            for reference in sorted(references, key=attrgetter("prefix", "identifier")):
                s = f"relationship: {typedef.curie} {reference.curie}"
                if typedef.name or reference.name:
                    s += " !"
                if typedef.name:
                    s += f" {typedef.name}"
                if reference.name:
                    s += f" {reference.name}"
                yield s

        for prop, value in sorted(self.iterate_properties()):
            yield f'property_value: {prop} "{value}" xsd:string'  # TODO deal with types later

        for synonym in sorted(self.synonyms, key=attrgetter("name")):
            yield synonym.to_obo()

    @staticmethod
    def _escape(s) -> str:
        return s.replace("\n", "\\n").replace('"', '\\"')


def _sort_relations(r):
    typedef, _references = r
    return typedef.reference.name or typedef.reference.identifier


class BioregistryError(ValueError):
    def __str__(self) -> str:
        return dedent(
            f"""
        The value you gave for Obo.ontology field ({self.args[0]}) is not a canonical
        Bioregistry prefix in the Obo.ontology field.

        Please see https://bioregistry.io for valid prefixes or feel free to open an issue
        on the PyOBO issue tracker for support.
        """
        )


@dataclass
class Obo:
    """An OBO document."""

    #: The prefix for the ontology
    ontology: ClassVar[str]

    #: Should the prefix be validated against the Bioregistry?
    check_bioregistry_prefix: ClassVar[bool] = True

    #: The name of the ontology. If not given, tries looking up with the Bioregistry.
    name: ClassVar[Optional[str]] = None

    #: The OBO format
    format_version: ClassVar[str] = "1.2"

    #: Type definitions
    typedefs: ClassVar[Optional[List[TypeDef]]] = None

    #: Synonym type definitions
    synonym_typedefs: ClassVar[Optional[List[SynonymTypeDef]]] = None

    #: An annotation about how an ontology was generated
    auto_generated_by: ClassVar[Optional[str]] = None

    #: The idspaces used in the document
    idspaces: ClassVar[Optional[Mapping[str, str]]] = None

    #: For super-sized datasets that shouldn't be read into memory
    iter_only: ClassVar[bool] = False

    #: Set to true for resources that are unversioned/very dynamic, like MGI
    dynamic_version: ClassVar[bool] = False

    #: Set to a static version for the resource (i.e., the resource is not itself versioned)
    static_version: ClassVar[Optional[str]] = None

    bioversions_key: ClassVar[Optional[str]] = None

    #: Root terms to use for the ontology
    root_terms: ClassVar[Optional[List[Reference]]] = None

    #: The date the ontology was generated
    date: Optional[datetime] = field(default_factory=datetime.today)

    #: The ontology version
    data_version: Optional[str] = None

    #: Should this ontology be reloaded?
    force: bool = False

    #: The hierarchy of terms
    _hierarchy: Optional[nx.DiGraph] = field(init=False, default=None, repr=False)
    #: A cache of terms
    _items: Optional[List[Term]] = field(init=False, default=None, repr=False)

    term_sort_key: ClassVar[Optional[Callable[["Obo", Term], int]]] = None

    def __post_init__(self):
        """Run post-init checks."""
        if self.ontology is None:
            raise ValueError
        if self.check_bioregistry_prefix and self.ontology != bioregistry.normalize_prefix(
            self.ontology
        ):
            raise BioregistryError(self.ontology)
        # The type ignores are because of the hack where we override the
        # class variables in the instance
        if self.name is None:
            self.name = bioregistry.get_name(self.ontology)  # type:ignore
        if not self.data_version:
            if self.static_version:
                self.data_version = self.static_version
            else:
                self.data_version = self._get_version()
        if not self.dynamic_version:
            if self.data_version is None:
                raise ValueError(f"{self.ontology} is missing data_version")
            elif "/" in self.data_version:
                raise ValueError(f"{self.ontology} has a slash in version: {self.data_version}")
        if self.auto_generated_by is None:
            self.auto_generated_by = f"bio2obo:{self.ontology}"  # type:ignore

    def _get_version(self) -> Optional[str]:
        if self.bioversions_key:
            import bioversions

            try:
                return bioversions.get_version(self.bioversions_key)
            except KeyError:
                logger.warning(f"[{self.bioversions_key}] bioversions doesn't list this resource ")
            except IOError:
                logger.warning(f"[{self.bioversions_key}] error while looking up version")
        return None

    @property
    def _version_or_raise(self) -> str:
        if not self.data_version:
            raise ValueError(f"There is no version available for {self.ontology}")
        return self.data_version

    def iter_terms(self, force: bool = False) -> Iterable[Term]:
        """Iterate over terms in this ontology."""
        raise NotImplementedError

    def get_graph(self):
        """Get an OBO Graph object."""
        from ..obographs import graph_from_obo

        return graph_from_obo(self)

    @classmethod
    def cli(cls) -> None:
        """Run the CLI for this class."""
        cli = cls.get_cls_cli()
        cli()

    @classmethod
    def get_cls_cli(cls) -> click.Command:
        """Get the CLI for this class."""

        @click.command()
        @verbose_option
        @force_option
        @click.option("--owl", is_flag=True, help="Write OWL via ROBOT")
        @click.option("--graph", is_flag=True, help="Write OBO Graph JSON via ROBOT")
        @click.option("--nodes", is_flag=True, help="Write nodes file")
        @click.option(
            "--version", help="Specify data version to get. Use this if bioversions is acting up."
        )
        def _main(force: bool, owl: bool, graph: bool, nodes: bool, version: Optional[str]):
            inst = cls(force=force, data_version=version)
            inst.write_default(
                write_obograph=graph,
                write_obo=True,
                write_owl=owl,
                write_nodes=nodes,
                force=force,
                use_tqdm=True,
            )

        return _main

    @property
    def date_formatted(self) -> str:
        """Get the date as a formatted string."""
        return (self.date if self.date else datetime.now()).strftime(DATE_FORMAT)

    def _iter_terms(self, use_tqdm: bool = False, desc: str = "terms") -> Iterable[Term]:
        if use_tqdm:
            total: Optional[int]
            try:
                total = len(self._items_accessor)
            except TypeError:
                total = None
            yield from tqdm(self, desc=desc, unit_scale=True, unit="term", total=total)
        else:
            yield from self

    def iterate_obo_lines(self) -> Iterable[str]:
        """Iterate over the lines to write in an OBO file."""
        yield f"format-version: {self.format_version}"
        yield f"date: {self.date_formatted}"

        if self.auto_generated_by is not None:
            yield f"auto-generated-by: {self.auto_generated_by}"

        if self.data_version is not None:
            yield f"data-version: {self.data_version}"

        for prefix, url in sorted((self.idspaces or {}).items()):
            yield f"idspace: {prefix} {url}"

        for synonym_typedef in sorted((self.synonym_typedefs or []), key=attrgetter("curie")):
            yield synonym_typedef.to_obo()

        yield f"ontology: {self.ontology}"

        if self.name is None:
            raise ValueError("ontology is missing name")
        yield f"remark: {self.name}"

        for typedef in sorted(self.typedefs or [], key=attrgetter("curie")):
            yield from typedef.iterate_obo_lines()

        for term in self:
            yield from term.iterate_obo_lines()

    def write_obo(
        self, file: Union[None, str, TextIO, Path] = None, use_tqdm: bool = False
    ) -> None:
        """Write the OBO to a file."""
        it = self.iterate_obo_lines()
        if use_tqdm:
            it = tqdm(it, desc=f"Writing {self.ontology}", unit_scale=True, unit="line")
        if isinstance(file, (str, Path, os.PathLike)):
            with open(file, "w") as fh:
                self._write_lines(it, fh)
        else:
            self._write_lines(it, file)

    @staticmethod
    def _write_lines(it, file: Optional[TextIO]):
        for line in it:
            print(line, file=file)  # noqa: T201

    def write_obonet_gz(self, path: Union[str, Path]) -> None:
        """Write the OBO to a gzipped dump in Obonet JSON."""
        graph = self.to_obonet()
        with gzip.open(path, "wt") as file:
            json.dump(nx.node_link_data(graph), file)

    def _path(self, *parts: str, name: Optional[str] = None) -> Path:
        return prefix_directory_join(self.ontology, *parts, name=name, version=self.data_version)

    def _cache(self, *parts: str, name: Optional[str] = None) -> Path:
        return self._path("cache", *parts, name=name)

    @property
    def _names_path(self) -> Path:
        return self._cache(name="names.tsv")

    @property
    def _definitions_path(self) -> Path:
        return self._cache(name="definitions.tsv")

    @property
    def _species_path(self) -> Path:
        return self._cache(name="species.tsv")

    @property
    def _synonyms_path(self) -> Path:
        return self._cache(name="synonyms.tsv")

    @property
    def _alts_path(self):
        return self._cache(name="alt_ids.tsv")

    @property
    def _typedefs_path(self) -> Path:
        return self._cache(name="typedefs.tsv")

    @property
    def _xrefs_path(self) -> Path:
        return self._cache(name="xrefs.tsv")

    @property
    def _relations_path(self) -> Path:
        return self._cache(name="relations.tsv")

    @property
    def _properties_path(self) -> Path:
        return self._cache(name="properties.tsv")

    @property
    def _root_metadata_path(self) -> Path:
        return prefix_directory_join(self.ontology, name="metadata.json")

    @property
    def _versioned_metadata_path(self) -> Path:
        return self._cache(name="metadata.json")

    @property
    def _obo_path(self) -> Path:
        return get_prefix_obo_path(self.ontology, version=self.data_version)

    @property
    def _obograph_path(self) -> Path:
        return self._path(name=f"{self.ontology}.json")

    @property
    def _owl_path(self) -> Path:
        return self._path(name=f"{self.ontology}.owl")

    @property
    def _obonet_gz_path(self) -> Path:
        return self._path(name=f"{self.ontology}.obonet.json.gz")

    @property
    def _nodes_path(self) -> Path:
        return self._path(name=f"{self.ontology}.nodes.tsv")

    def write_default(
        self,
        use_tqdm: bool = False,
        force: bool = False,
        write_obo: bool = False,
        write_obonet: bool = False,
        write_obograph: bool = False,
        write_owl: bool = False,
        write_nodes: bool = False,
    ) -> None:
        """Write the OBO to the default path."""
        metadata = self.get_metadata()
        for path in (self._root_metadata_path, self._versioned_metadata_path):
            logger.debug("[%s v%s] caching metadata to %s", self.ontology, self.data_version, path)
            with path.open("w") as file:
                json.dump(metadata, file, indent=2)

        logger.debug(
            "[%s v%s] caching typedefs to %s", self.ontology, self.data_version, self._typedefs_path
        )
        typedef_df: pd.DataFrame = self.get_typedef_df()
        typedef_df.sort_values(list(typedef_df.columns), inplace=True)
        typedef_df.to_csv(self._typedefs_path, sep="\t", index=False)

        for label, path, header, fn in [
            ("names", self._names_path, [f"{self.ontology}_id", "name"], self.iterate_id_name),
            (
                "definitions",
                self._definitions_path,
                [f"{self.ontology}_id", "definition"],
                self.iterate_id_definition,
            ),
            (
                "species",
                self._species_path,
                [f"{self.ontology}_id", "taxonomy_id"],
                self.iterate_id_species,
            ),
            (
                "synonyms",
                self._synonyms_path,
                [f"{self.ontology}_id", "synonym"],
                self.iterate_synonym_rows,
            ),
            ("alts", self._alts_path, [f"{self.ontology}_id", "alt_id"], self.iterate_alt_rows),
            ("xrefs", self._xrefs_path, self.xrefs_header, self.iterate_xref_rows),
            ("relations", self._relations_path, self.relations_header, self.iter_relation_rows),
            ("properties", self._properties_path, self.properties_header, self.iter_property_rows),
        ]:
            if path.exists() and not force:
                continue
            logger.debug("[%s v%s] caching %s to %s", self.ontology, self.data_version, label, path)
            write_iterable_tsv(
                path=path,
                header=header,
                it=fn(),  # type:ignore
            )

        for relation in (is_a, has_part, part_of, from_species, orthologous):
            if relation is not is_a and self.typedefs is not None and relation not in self.typedefs:
                continue
            relations_path = self._cache("relations", name=f"{relation.curie}.tsv")
            if relations_path.exists() and not force:
                continue
            logger.debug(
                "[%s v%s] caching relation %s ! %s",
                self.ontology,
                self.data_version,
                relation.curie,
                relation.name,
            )
            relation_df = self.get_filtered_relations_df(relation)
            if not len(relation_df.index):
                continue
            relation_df.sort_values(list(relation_df.columns), inplace=True)
            relation_df.to_csv(relations_path, sep="\t", index=False)

        if (write_obo or write_obograph or write_owl) and (not self._obo_path.exists() or force):
            self.write_obo(self._obo_path, use_tqdm=use_tqdm)
        if write_obograph:
            # obo_to_obograph(self._obo_path, self._obograph_path)
            self._obograph_path.write_text(
                self.get_graph().json(
                    indent=2, ensure_ascii=False, exclude_none=True, exclude_unset=True
                )
            )
        if write_owl:
            obo_to_owl(self._obo_path, self._owl_path)
        if write_obonet and (not self._obonet_gz_path.exists() or force):
            logger.debug("writing obonet to %s", self._obonet_gz_path)
            self.write_obonet_gz(self._obonet_gz_path)
        if write_nodes:
            self.get_graph().get_nodes_df().to_csv(self._nodes_path, sep="\t", index=False)

    @property
    def _items_accessor(self):
        if self._items is None:
            key = self.term_sort_key or attrgetter("curie")
            self._items = sorted(self.iter_terms(force=self.force), key=key)
        return self._items

    def __iter__(self) -> Iterator["Term"]:  # noqa: D105
        if self.iter_only:
            return iter(self.iter_terms(force=self.force))
        return iter(self._items_accessor)

    def ancestors(self, identifier: str) -> Set[str]:
        """Return a set of identifiers for parents of the given identifier."""
        return nx.descendants(self.hierarchy, identifier)  # note this is backwards

    def descendants(self, identifier: str) -> Set[str]:
        """Return a set of identifiers for the children of the given identifier."""
        return nx.ancestors(self.hierarchy, identifier)  # note this is backwards

    def is_descendant(self, descendant: str, ancestor: str) -> bool:
        """Return if the given identifier is a descendent of the ancestor.

        .. code-block:: python

            from pyobo import get_obo
            obo = get_obo('go')

            interleukin_10_complex = '1905571'  # interleukin-10 receptor complex
            all_complexes = '0032991'
            assert obo.is_descendant('1905571', '0032991')
        """
        return ancestor in self.ancestors(descendant)

    @property
    def hierarchy(self) -> nx.DiGraph:
        """A graph representing the parent/child relationships between the entities.

        To get all children of a given entity, do:

        .. code-block:: python

            from pyobo import get_obo
            obo = get_obo('go')

            identifier = '1905571'  # interleukin-10 receptor complex
            is_complex = '0032991' in  nx.descendants(obo.hierarchy, identifier)  # should be true
        """  # noqa:D401
        if self._hierarchy is None:
            self._hierarchy = nx.DiGraph()
            for term in self._iter_terms(desc=f"[{self.ontology}] getting hierarchy"):
                for parent in term.parents:
                    self._hierarchy.add_edge(term.identifier, parent.identifier)
        return self._hierarchy

    def to_obonet(self: "Obo", *, use_tqdm: bool = False) -> nx.MultiDiGraph:
        """Export as a :mod`obonet` style graph."""
        rv = nx.MultiDiGraph()
        rv.graph.update(
            {
                "name": self.name,
                "ontology": self.ontology,
                "auto-generated-by": self.auto_generated_by,
                "typedefs": _convert_typedefs(self.typedefs),
                "format-version": self.format_version,
                "data-version": self.data_version,
                "synonymtypedef": _convert_synonym_typedefs(self.synonym_typedefs),
                "date": self.date_formatted,
            }
        )

        nodes = {}
        links = []
        for term in self._iter_terms(use_tqdm=use_tqdm):
            parents = []
            for parent in term.parents:
                if parent is None:
                    raise ValueError("parent should not be none!")
                links.append((term.curie, "is_a", parent.curie))
                parents.append(parent.curie)

            relations = []
            for typedef, target in term.iterate_relations():
                if target is None:
                    raise ValueError("target should not be none!")
                relations.append(f"{typedef.curie} {target.curie}")
                links.append((term.curie, typedef.curie, target.curie))

            d = {
                "id": term.curie,
                "name": term.name,
                "def": term.definition and term._definition_fp(),
                "xref": [xref.curie for xref in term.xrefs],
                "is_a": parents,
                "relationship": relations,
                "synonym": [synonym._fp() for synonym in term.synonyms],
                "property_value": [
                    f"{prop} {value}"
                    for prop, values in term.properties.items()
                    for value in values
                ],
            }
            nodes[term.curie] = {k: v for k, v in d.items() if v}

        rv.add_nodes_from(nodes.items())
        for _source, _key, _target in links:
            rv.add_edge(_source, _target, key=_key)

        logger.info(
            "[%s v%s] exported graph with %d nodes",
            self.ontology,
            self.data_version,
            rv.number_of_nodes(),
        )
        return rv

    def get_metadata(self) -> Mapping[str, Any]:
        """Get metadata."""
        return dict(
            version=self.data_version,
            date=self.date and self.date.isoformat(),
        )

    def iterate_ids(self, *, use_tqdm: bool = False) -> Iterable[str]:
        """Iterate over identifiers."""
        for term in self._iter_terms(use_tqdm=use_tqdm, desc=f"[{self.ontology}] getting names"):
            if term.prefix == self.ontology:
                yield term.identifier

    def get_ids(self, *, use_tqdm: bool = False) -> Set[str]:
        """Get the set of identifiers."""
        return set(self.iterate_ids(use_tqdm=use_tqdm))

    def iterate_id_name(self, *, use_tqdm: bool = False) -> Iterable[Tuple[str, str]]:
        """Iterate identifier name pairs."""
        for term in self._iter_terms(use_tqdm=use_tqdm, desc=f"[{self.ontology}] getting names"):
            if term.name:
                yield term.identifier, term.name

    def get_id_name_mapping(self, *, use_tqdm: bool = False) -> Mapping[str, str]:
        """Get a mapping from identifiers to names."""
        return dict(self.iterate_id_name(use_tqdm=use_tqdm))

    def iterate_id_definition(self, *, use_tqdm: bool = False) -> Iterable[Tuple[str, str]]:
        """Iterate over pairs of terms' identifiers and their respective definitions."""
        for term in self._iter_terms(use_tqdm=use_tqdm, desc=f"[{self.ontology}] getting names"):
            if term.identifier and term.definition:
                yield term.identifier, term.definition.strip('"').replace("\n", " ").replace(
                    "\t", " "
                ).replace("  ", " ")

    def get_id_definition_mapping(self, *, use_tqdm: bool = False) -> Mapping[str, str]:
        """Get a mapping from identifiers to definitions."""
        return dict(self.iterate_id_definition(use_tqdm=use_tqdm))

    ############
    # TYPEDEFS #
    ############

    def iterate_id_species(
        self, *, prefix: Optional[str] = None, use_tqdm: bool = False
    ) -> Iterable[Tuple[str, str]]:
        """Iterate over terms' identifiers and respective species (if available)."""
        if prefix is None:
            prefix = NCBITAXON_PREFIX
        for term in self._iter_terms(use_tqdm=use_tqdm, desc=f"[{self.ontology}] getting species"):
            species = term.get_species(prefix=prefix)
            if species:
                yield term.identifier, species.identifier

    def get_id_species_mapping(
        self, *, prefix: Optional[str] = None, use_tqdm: bool = False
    ) -> Mapping[str, str]:
        """Get a mapping from identifiers to species."""
        return dict(self.iterate_id_species(prefix=prefix, use_tqdm=use_tqdm))

    ############
    # TYPEDEFS #
    ############

    def get_typedef_df(self, use_tqdm: bool = False) -> pd.DataFrame:
        """Get a typedef dataframe."""
        rows = [
            (typedef.prefix, typedef.identifier, typedef.name)
            for typedef in tqdm(self.typedefs or [], disable=not use_tqdm)
        ]
        return pd.DataFrame(rows, columns=["prefix", "identifier", "name"])

    def iter_typedef_id_name(self) -> Iterable[Tuple[str, str]]:
        """Iterate over typedefs' identifiers and their respective names."""
        for typedef in self.typedefs or []:
            yield typedef.identifier, typedef.name

    def get_typedef_id_name_mapping(self) -> Mapping[str, str]:
        """Get a mapping from typedefs' identifiers to names."""
        return dict(self.iter_typedef_id_name())

    #########
    # PROPS #
    #########

    def iterate_properties(self, *, use_tqdm: bool = False) -> Iterable[Tuple[Term, str, str]]:
        """Iterate over tuples of terms, properties, and their values."""
        # TODO if property_prefix is set, try removing that as a prefix from all prop strings.
        for term in self._iter_terms(
            use_tqdm=use_tqdm, desc=f"[{self.ontology}] getting properties"
        ):
            for prop, value in term.iterate_properties():
                yield term, prop, value

    @property
    def properties_header(self):
        """Property dataframe header."""  # noqa:D401
        return [f"{self.ontology}_id", "property", "value"]

    def iter_property_rows(self, *, use_tqdm: bool = False) -> Iterable[Tuple[str, str, str]]:
        """Iterate property rows."""
        for term, prop, value in self.iterate_properties(use_tqdm=use_tqdm):
            yield term.identifier, prop, value

    def get_properties_df(self, *, use_tqdm: bool = False) -> pd.DataFrame:
        """Get all properties as a dataframe."""
        return pd.DataFrame(
            list(self.iter_property_rows(use_tqdm=use_tqdm)),
            columns=self.properties_header,
        )

    def iterate_filtered_properties(
        self, prop: str, *, use_tqdm: bool = False
    ) -> Iterable[Tuple[Term, str]]:
        """Iterate over tuples of terms and the values for the given property."""
        for term in self._iter_terms(use_tqdm=use_tqdm):
            for _prop, value in term.iterate_properties():
                if _prop == prop:
                    yield term, value

    def get_filtered_properties_df(self, prop: str, *, use_tqdm: bool = False) -> pd.DataFrame:
        """Get a dataframe of terms' identifiers to the given property's values."""
        return pd.DataFrame(
            list(self.get_filtered_properties_mapping(prop, use_tqdm=use_tqdm).items()),
            columns=[f"{self.ontology}_id", prop],
        )

    def get_filtered_properties_mapping(
        self, prop: str, *, use_tqdm: bool = False
    ) -> Mapping[str, str]:
        """Get a mapping from a term's identifier to the property.

        .. warning:: Assumes there's only one version of the property for each term.
        """
        return {
            term.identifier: value
            for term, value in self.iterate_filtered_properties(prop, use_tqdm=use_tqdm)
        }

    def get_filtered_properties_multimapping(
        self, prop: str, *, use_tqdm: bool = False
    ) -> Mapping[str, List[str]]:
        """Get a mapping from a term's identifier to the property values."""
        return multidict(
            (term.identifier, value)
            for term, value in self.iterate_filtered_properties(prop, use_tqdm=use_tqdm)
        )

    #############
    # RELATIONS #
    #############

    def iterate_relations(
        self, *, use_tqdm: bool = False
    ) -> Iterable[Tuple[Term, TypeDef, Reference]]:
        """Iterate over tuples of terms, relations, and their targets."""
        for term in self._iter_terms(
            use_tqdm=use_tqdm, desc=f"[{self.ontology}] getting relations"
        ):
            for parent in term.parents:
                yield term, is_a, parent
            for typedef, reference in term.iterate_relations():
                if (self.typedefs is None or typedef not in self.typedefs) and (
                    typedef.prefix,
                    typedef.identifier,
                ) not in default_typedefs:
                    raise ValueError(f"Undefined typedef: {typedef.curie} ! {typedef.name}")
                yield term, typedef, reference

    def iter_relation_rows(
        self, use_tqdm: bool = False
    ) -> Iterable[Tuple[str, str, str, str, str]]:
        """Iterate the relations' rows."""
        for term, typedef, reference in self.iterate_relations(use_tqdm=use_tqdm):
            yield term.identifier, typedef.prefix, typedef.identifier, reference.prefix, reference.identifier

    def iterate_filtered_relations(
        self,
        relation: RelationHint,
        *,
        use_tqdm: bool = False,
    ) -> Iterable[Tuple[Term, Reference]]:
        """Iterate over tuples of terms and ther targets for the given relation."""
        _target_prefix, _target_identifier = get_reference_tuple(relation)
        for term, typedef, reference in self.iterate_relations(use_tqdm=use_tqdm):
            if typedef.prefix == _target_prefix and typedef.identifier == _target_identifier:
                yield term, reference

    @property
    def relations_header(self) -> Sequence[str]:
        """Header for the relations dataframe."""  # noqa:D401
        return [f"{self.ontology}_id", RELATION_PREFIX, RELATION_ID, TARGET_PREFIX, TARGET_ID]

    def get_relations_df(self, *, use_tqdm: bool = False) -> pd.DataFrame:
        """Get all relations from the OBO."""
        return pd.DataFrame(
            list(self.iter_relation_rows(use_tqdm=use_tqdm)),
            columns=self.relations_header,
        )

    def get_filtered_relations_df(
        self,
        relation: RelationHint,
        *,
        use_tqdm: bool = False,
    ) -> pd.DataFrame:
        """Get a specific relation from OBO."""
        return pd.DataFrame(
            [
                (term.identifier, reference.prefix, reference.identifier)
                for term, reference in self.iterate_filtered_relations(relation, use_tqdm=use_tqdm)
            ],
            columns=[f"{self.ontology}_id", TARGET_PREFIX, TARGET_ID],
        )

    def iterate_filtered_relations_filtered_targets(
        self,
        relation: RelationHint,
        target_prefix: str,
        *,
        use_tqdm: bool = False,
    ) -> Iterable[Tuple[Term, Reference]]:
        """Iterate over relationships between one identifier and another."""
        for term, reference in self.iterate_filtered_relations(
            relation=relation, use_tqdm=use_tqdm
        ):
            if reference.prefix == target_prefix:
                yield term, reference

    def get_relation_mapping(
        self,
        relation: RelationHint,
        target_prefix: str,
        *,
        use_tqdm: bool = False,
    ) -> Mapping[str, str]:
        """Get a mapping from the term's identifier to the target's identifier.

        .. warning:: Assumes there's only one version of the property for each term.

         Example usage: get homology between HGNC and MGI:

        >>> from pyobo.sources.hgnc import get_obo
        >>> obo = get_obo()
        >>> human_mapt_hgnc_id = '6893'
        >>> mouse_mapt_mgi_id = '97180'
        >>> hgnc_mgi_orthology_mapping = obo.get_relation_mapping('ro:HOM0000017', 'mgi')
        >>> assert mouse_mapt_mgi_id == hgnc_mgi_orthology_mapping[human_mapt_hgnc_id]
        """
        return {
            term.identifier: reference.identifier
            for term, reference in self.iterate_filtered_relations_filtered_targets(
                relation=relation,
                target_prefix=target_prefix,
                use_tqdm=use_tqdm,
            )
        }

    def get_relation(
        self,
        source_identifier: str,
        relation: RelationHint,
        target_prefix: str,
        *,
        use_tqdm: bool = False,
    ) -> Optional[str]:
        """Get the value for a bijective relation mapping between this resource and a target resource.

        >>> from pyobo.sources.hgnc import get_obo
        >>> obo = get_obo()
        >>> human_mapt_hgnc_id = '6893'
        >>> mouse_mapt_mgi_id = '97180'
        >>> assert mouse_mapt_mgi_id == obo.get_relation(human_mapt_hgnc_id, 'ro:HOM0000017', 'mgi')
        """
        relation_mapping = self.get_relation_mapping(
            relation=relation, target_prefix=target_prefix, use_tqdm=use_tqdm
        )
        return relation_mapping.get(source_identifier)

    def get_relation_multimapping(
        self,
        relation: RelationHint,
        target_prefix: str,
        *,
        use_tqdm: bool = False,
    ) -> Mapping[str, List[str]]:
        """Get a mapping from the term's identifier to the target's identifiers."""
        return multidict(
            (term.identifier, reference.identifier)
            for term, reference in self.iterate_filtered_relations_filtered_targets(
                relation=relation,
                target_prefix=target_prefix,
                use_tqdm=use_tqdm,
            )
        )

    def get_id_multirelations_mapping(
        self,
        typedef: TypeDef,
        *,
        use_tqdm: bool = False,
    ) -> Mapping[str, List[Reference]]:
        """Get a mapping from identifiers to a list of all references for the given relation."""
        return multidict(
            (term.identifier, reference)
            for term in self._iter_terms(
                use_tqdm=use_tqdm, desc=f"[{self.ontology}] getting {typedef.curie}"
            )
            for reference in term.get_relationships(typedef)
        )

    ############
    # SYNONYMS #
    ############

    def iterate_synonyms(self, *, use_tqdm: bool = False) -> Iterable[Tuple[Term, Synonym]]:
        """Iterate over pairs of term and synonym object."""
        for term in self._iter_terms(use_tqdm=use_tqdm, desc=f"[{self.ontology}] getting synonyms"):
            for synonym in sorted(term.synonyms, key=attrgetter("name")):
                yield term, synonym

    def iterate_synonym_rows(self, *, use_tqdm: bool = False) -> Iterable[Tuple[str, str]]:
        """Iterate over pairs of identifier and synonym text."""
        for term, synonym in self.iterate_synonyms(use_tqdm=use_tqdm):
            yield term.identifier, synonym.name

    def get_id_synonyms_mapping(self, *, use_tqdm: bool = False) -> Mapping[str, List[str]]:
        """Get a mapping from identifiers to a list of sorted synonym strings."""
        return multidict(self.iterate_synonym_rows(use_tqdm=use_tqdm))

    #########
    # XREFS #
    #########

    def iterate_xrefs(self, *, use_tqdm: bool = False) -> Iterable[Tuple[Term, Reference]]:
        """Iterate over xrefs."""
        for term in self._iter_terms(use_tqdm=use_tqdm, desc=f"[{self.ontology}] getting xrefs"):
            for xref in term.xrefs:
                yield term, xref

    def iterate_filtered_xrefs(
        self, prefix: str, *, use_tqdm: bool = False
    ) -> Iterable[Tuple[Term, Reference]]:
        """Iterate over xrefs to a given prefix."""
        for term, xref in self.iterate_xrefs(use_tqdm=use_tqdm):
            if xref.prefix == prefix:
                yield term, xref

    def iterate_xref_rows(self, *, use_tqdm: bool = False) -> Iterable[Tuple[str, str, str]]:
        """Iterate over terms' identifiers, xref prefixes, and xref identifiers."""
        for term, xref in self.iterate_xrefs(use_tqdm=use_tqdm):
            yield term.identifier, xref.prefix, xref.identifier

    @property
    def xrefs_header(self):
        """The header for the xref dataframe."""  # noqa:D401
        return [f"{self.ontology}_id", TARGET_PREFIX, TARGET_ID]

    def get_xrefs_df(self, *, use_tqdm: bool = False) -> pd.DataFrame:
        """Get a dataframe of all xrefs extracted from the OBO document."""
        return pd.DataFrame(
            list(self.iterate_xref_rows(use_tqdm=use_tqdm)),
            columns=[f"{self.ontology}_id", TARGET_PREFIX, TARGET_ID],
        ).drop_duplicates()

    def get_filtered_xrefs_mapping(
        self, prefix: str, *, use_tqdm: bool = False
    ) -> Mapping[str, str]:
        """Get filtered xrefs as a dictionary."""
        return {
            term.identifier: xref.identifier
            for term, xref in self.iterate_filtered_xrefs(prefix, use_tqdm=use_tqdm)
        }

    def get_filtered_multixrefs_mapping(
        self, prefix: str, *, use_tqdm: bool = False
    ) -> Mapping[str, List[str]]:
        """Get filtered xrefs as a dictionary."""
        return multidict(
            (term.identifier, xref.identifier)
            for term, xref in self.iterate_filtered_xrefs(prefix, use_tqdm=use_tqdm)
        )

    ########
    # ALTS #
    ########

    def iterate_alts(self) -> Iterable[Tuple[Term, Reference]]:
        """Iterate over alternative identifiers."""
        for term in self:
            for alt in term.alt_ids:
                yield term, alt

    def iterate_alt_rows(self) -> Iterable[Tuple[str, str]]:
        """Iterate over pairs of terms' primary identifiers and alternate identifiers."""
        for term, alt in self.iterate_alts():
            yield term.identifier, alt.identifier

    def get_id_alts_mapping(self) -> Mapping[str, List[str]]:
        """Get a mapping from identifiers to a list of alternative identifiers."""
        return multidict((term.identifier, alt.identifier) for term, alt in self.iterate_alts())


def make_ad_hoc_ontology(
    _ontology: str,
    _name: str,
    _auto_generated_by: Optional[str] = None,
    _format_version: str = "1.2",
    _typedefs: Optional[List[TypeDef]] = None,
    _synonym_typedefs: Optional[List[SynonymTypeDef]] = None,
    _date: Optional[datetime] = None,
    _data_version: Optional[str] = None,
    _idspaces: Optional[Mapping[str, str]] = None,
    *,
    terms: List[Term],
) -> "Obo":
    """Make an ad-hoc ontology."""

    class AdHocOntology(Obo):
        """An ad hoc ontology created from an OBO file."""

        ontology = _ontology
        name = _name
        auto_generated_by = _auto_generated_by
        format_version = _format_version
        typedefs = _typedefs
        synonym_typedefs = _synonym_typedefs
        idspaces = _idspaces

        def __post_init__(self):
            self.date = _date
            self.data_version = _data_version

        def iter_terms(self, force: bool = False) -> Iterable[Term]:
            """Iterate over terms in the ad hoc ontology."""
            return terms

    return AdHocOntology()


def _convert_typedefs(typedefs: Optional[Iterable[TypeDef]]) -> List[Mapping[str, Any]]:
    """Convert the type defs."""
    if not typedefs:
        return []
    return [_convert_typedef(typedef) for typedef in typedefs]


def _convert_typedef(typedef: TypeDef) -> Mapping[str, Any]:
    """Convert a type def."""
    # TODO add more later
    return typedef.reference.dict()


def _convert_synonym_typedefs(synonym_typedefs: Optional[Iterable[SynonymTypeDef]]) -> List[str]:
    """Convert the synonym type defs."""
    if not synonym_typedefs:
        return []
    return [_convert_synonym_typedef(synonym_typedef) for synonym_typedef in synonym_typedefs]


def _convert_synonym_typedef(synonym_typedef: SynonymTypeDef) -> str:
    return f'{synonym_typedef.curie} "{synonym_typedef.name}"'



# -*- coding: utf-8 -*-

"""Test mapping program."""

import unittest

import pandas as pd

from pyobo import Canonicalizer
from pyobo.constants import XREF_COLUMNS
from pyobo.xrefdb.xrefs_pipeline import get_graph_from_xref_df


class TestCanonicalizer(unittest.TestCase):
    """Tests for the canonicalizer."""

    canonicalizer: Canonicalizer

    def setUp(self) -> None:
        """Set up a test dataframe for canonicalization."""
        df = pd.DataFrame(
            [
                # Cluster 1 - fully connected
                ["hgnc", "h1", "ensembl", "e1", "example_source"],
                ["hgnc", "h1", "omim", "o1", "example_source"],
                ["omim", "o1", "cds", "c1", "example_source"],
                # Cluster 2 - missing HGNC
                ["omim", "o2", "ensembl", "e2"],
                # Cluster 3 - irrelevant
                ["y", "y1", "z", "z1", "irrelevant_source"],
            ],
            columns=XREF_COLUMNS,
        )
        graph = get_graph_from_xref_df(df)
        priority = [
            "hgnc",
            "omim",
            "ensembl",
        ]
        self.canonicalizer = Canonicalizer(graph=graph, priority=priority)

    def test_priority(self):
        """Test lookup priority."""
        self.assertEqual(3, self.canonicalizer._key("hgnc:h1"))
        self.assertEqual(2, self.canonicalizer._key("omim:o1"))
        self.assertEqual(1, self.canonicalizer._key("ensembl:e1"))
        # Since CDS isn't there, it gets no priority at all
        self.assertEqual(None, self.canonicalizer._key("cds:c1"))

    def test_mapper(self):
        """Test mapping back to hgnc for cluster 1."""
        for curie in ["hgnc:h1", "ensembl:e1", "omim:o1", "cds:c1"]:
            with self.subTest(curie=curie):
                self.assertEqual("hgnc:h1", self.canonicalizer.canonicalize(curie))

        # Test a non-present node
        self.assertEqual("xxx:x1", self.canonicalizer.canonicalize("xxx:x1"))

        for curie in ["omim:o2", "ensembl:e2"]:
            with self.subTest(curie=curie):
                self.assertEqual("omim:o2", self.canonicalizer.canonicalize(curie))

        # Test a present node with no priority information
        self.assertEqual("y:y1", self.canonicalizer.canonicalize("y:y1"))
        self.assertEqual("z:z1", self.canonicalizer.canonicalize("z:z1"))



# -*- coding: utf-8 -*-

"""Converter for Entrez."""

import logging
from typing import Iterable, List, Mapping, Set

import bioregistry
import pandas as pd
from tqdm.auto import tqdm

from ..struct import Obo, Reference, Term, from_species
from ..utils.path import ensure_df

__all__ = [
    "NCBIGeneGetter",
]

logger = logging.getLogger(__name__)

PREFIX = "ncbigene"

SPECIES_CONSORTIUM_MAPPING = {
    "10090": "mgi",  # Mouse
    "10116": "rgd",  # Rat
    "4932": "sgd",  # Yeast
    "7227": "flybase",  # Drosophila
    "9606": "hgnc",  # Human
    "6239": "wormbase",  # C. Elegans
    "7955": "zfin",  # Zebrafish
}

#: All namepace codes (in lowercase) that can map to ncbigene
CONSORTIUM_SPECIES_MAPPING = {
    database_code: taxonomy_id for taxonomy_id, database_code in SPECIES_CONSORTIUM_MAPPING.items()
}

GENE_INFO_URL = "ftp://ftp.ncbi.nlm.nih.gov/gene/DATA/gene_info.gz"
#: Columns fro gene_info.gz that are used
GENE_INFO_COLUMNS = [
    "#tax_id",
    "GeneID",
    "Symbol",
    "dbXrefs",
    "description",
    "type_of_gene",
]


def get_ncbigene_ids() -> Set[str]:
    """Get the Entrez name mapping."""
    df = _get_ncbigene_subset(["GeneID"])
    return set(df["GeneID"])


def get_ncbigene_id_to_name_mapping() -> Mapping[str, str]:
    """Get the Entrez name mapping."""
    return _get_ncbigene_info_subset(["GeneID", "Symbol"])


def get_ncbigene_id_to_species_mapping() -> Mapping[str, str]:
    """Get the Entrez species mapping."""
    return _get_ncbigene_info_subset(["GeneID", "Symbol"])


def _get_ncbigene_info_subset(usecols) -> Mapping[str, str]:
    df = _get_ncbigene_subset(usecols)
    return dict(df.values)


def _get_ncbigene_subset(usecols: List[str]) -> pd.DataFrame:
    df = ensure_df(
        PREFIX,
        url=GENE_INFO_URL,
        sep="\t",
        na_values=["-", "NEWENTRY"],
        usecols=usecols,
        dtype=str,
    )
    df.dropna(inplace=True)
    if len(usecols) > 1:
        df = df[usecols]
    return df


class NCBIGeneGetter(Obo):
    """An ontology representation of NCBI's Entrez Gene database."""

    ontology = PREFIX
    dynamic_version = True
    typedefs = [from_species]

    def iter_terms(self, force: bool = False) -> Iterable[Term]:
        """Iterate over terms in the ontology."""
        return get_terms(force=force)


def get_obo(force: bool = False) -> Obo:
    """Get Entrez as OBO."""
    return NCBIGeneGetter(force=force)


def get_gene_info_df(force: bool = False) -> pd.DataFrame:
    """Get the gene info dataframe."""
    return ensure_df(
        PREFIX,
        url=GENE_INFO_URL,
        sep="\t",
        na_values=["-", "NEWENTRY"],
        usecols=GENE_INFO_COLUMNS,
        dtype=str,
        force=force,
    )


"""xref_mapping was obtained from:

namespaces = set()
for xrefs in df[df['dbXrefs'].notna()]['dbXrefs']:
    for xref in xrefs.split('|'):
        namespaces.add(xref.split(':')[0])

print('namespaces:')
print(*sorted(namespaces), sep='\n')
"""

xref_mapping = {
    "APHIDBASE",
    "ASAP",
    "AnimalQTLdb",
    "ApiDB_CryptoDB",
    "Araport",
    "BEEBASE",
    "BEETLEBASE",
    "BGD",
    "CGNC",
    "ECOCYC",
    "EcoGene",
    "Ensembl",
    "FLYBASE",
    "HGNC",
    "IMGT/GENE-DB",
    "MGI",
    "MIM",
    "NASONIABASE",
    "PseudoCap",
    "RGD",
    "SGD",
    "TAIR",
    "VGNC",
    "VectorBase",
    "WormBase",
    "Xenbase",
    "ZFIN",
    "dictyBase",
    "miRBase",
}
xref_mapping = {x.lower() for x in xref_mapping}


def get_terms(force: bool = False) -> Iterable[Term]:
    """Get Entrez terms."""
    df = get_gene_info_df(force=force)

    it = tqdm(
        df.values, total=len(df.index), desc=f"mapping {PREFIX}", unit_scale=True, unit="gene"
    )
    warning_prefixes = set()
    for tax_id, gene_id, symbol, xref_curies, description, _gene_type in it:
        if pd.isna(symbol):
            continue
        term = Term(
            reference=Reference(prefix=PREFIX, identifier=gene_id, name=symbol),
            definition=description,
        )
        term.set_species(identifier=tax_id)
        if pd.notna(xref_curies):
            for xref_curie in xref_curies.split("|"):
                if xref_curie.startswith("EnsemblRapid"):
                    continue
                if xref_curie.startswith("AllianceGenome"):
                    xref_curie = xref_curie[len("xref_curie") :]
                xref_prefix, xref_id = bioregistry.parse_curie(xref_curie)
                if xref_prefix and xref_id:
                    term.append_xref(Reference(prefix=xref_prefix, identifier=xref_id))
                else:
                    p = xref_curie.split(":")[0]
                    if p not in warning_prefixes:
                        warning_prefixes.add(p)
                        tqdm.write(f"[{PREFIX}] unhandled xref prefix: {p}")
        yield term


if __name__ == "__main__":
    get_obo().write_default()



# -*- coding: utf-8 -*-

"""Tests for PyOBO."""

import unittest

import pandas as pd

import pyobo
from pyobo import get_filtered_xrefs, get_id_name_mapping, get_xrefs_df
from pyobo.constants import TARGET_ID, TARGET_PREFIX
from pyobo.mocks import get_mock_get_xrefs_df
from tests.constants import chebi_patch

_mock_rows = [
    ("hgnc", "6893", "ncbigene", "4137", "N/A"),
    ("hgnc", "6893", "ensembl", "ENSG00000186868", "N/A"),
]
mock_get_xrefs_df = get_mock_get_xrefs_df(_mock_rows)


class TestMapping(unittest.TestCase):
    """Test extracting information."""

    def test_get_names(self):
        """Test getting names."""
        with chebi_patch:
            id_to_name = get_id_name_mapping("chebi")
        for identifier in id_to_name:
            self.assertFalse(identifier.startswith("CHEBI"))
            self.assertFalse(identifier.startswith("CHEBI:"))
            self.assertFalse(identifier.startswith("chebi:"))
            self.assertFalse(identifier.startswith("chebi"))

    def test_get_xrefs(self):
        """Test getting xrefs."""
        with chebi_patch:
            df = get_xrefs_df("chebi")
        self.assertIsInstance(df, pd.DataFrame)
        self.assertEqual(["chebi_id", TARGET_PREFIX, TARGET_ID], list(df.columns))

    def test_get_target_xrefs(self):
        """Test getting xrefs."""
        with chebi_patch:
            kegg_xrefs = get_filtered_xrefs("chebi", "kegg")

        for key, value in kegg_xrefs.items():
            self.assertFalse(key.startswith("CHEBI:"))
            self.assertFalse(key.startswith("CHEBI"))
            self.assertFalse(key.startswith("chebi:"))
            self.assertFalse(key.startswith("chebi"))
            self.assertFalse(value.startswith("KEGG:"))
            self.assertFalse(value.startswith("KEGG"))
            self.assertFalse(value.startswith("kegg:"))
            self.assertFalse(value.startswith("kegg"))

        self.assertIsInstance(kegg_xrefs, dict)

    @mock_get_xrefs_df
    def test_get_equivalent(self, _):
        """Test getting equivalent CURIEs."""
        mapt_curies = pyobo.get_equivalent("hgnc:6893")
        self.assertIn("ncbigene:4137", mapt_curies)
        self.assertIn("ensembl:ENSG00000186868", mapt_curies)
        self.assertNotIn("hgnc:6893", mapt_curies)



# -*- coding: utf-8 -*-

"""PyOBO's Mapping Service.

Run with ``python -m pyobo.apps.mapper``.
"""

import logging
from functools import lru_cache
from typing import Any, Dict, Iterable, List, Mapping, Optional, Union

import bioregistry
import pandas as pd
from flasgger import Swagger
from flask import (
    Blueprint,
    Flask,
    abort,
    current_app,
    jsonify,
    render_template,
    url_for,
)
from flask_bootstrap import VERSION_BOOTSTRAP, Bootstrap
from werkzeug.local import LocalProxy

from pyobo import Canonicalizer
from pyobo.constants import PROVENANCE, SOURCE_PREFIX, TARGET_PREFIX
from pyobo.identifier_utils import normalize_curie
from pyobo.resource_utils import ensure_inspector_javert_df

__all__ = [
    "get_app",
]

logger = logging.getLogger(__name__)

summary_df = LocalProxy(lambda: current_app.config["summary"])
canonicalizer: Canonicalizer = LocalProxy(lambda: current_app.config["canonicalizer"])


@lru_cache()
def _single_source_shortest_path(curie: str) -> Optional[Mapping[str, List[Mapping[str, str]]]]:
    return canonicalizer.single_source_shortest_path(curie=curie)


@lru_cache()
def _all_shortest_paths(source_curie: str, target_curie: str) -> List[List[Mapping[str, str]]]:
    return canonicalizer.all_shortest_paths(source_curie=source_curie, target_curie=target_curie)


#: The blueprint that gets added to the app
search_blueprint = Blueprint("search", __name__)


@search_blueprint.route("/")
def home():
    """Show the home page."""
    return render_template("mapper_home.html")


@search_blueprint.route("/mappings/<curie>")
def single_source_mappings(curie: str):
    """Return all length xrefs from the given identifier."""
    if curie not in canonicalizer.graph:
        return jsonify(
            success=False,
            query=dict(curie=curie),
            message="could not find curie",
        )
    return jsonify(_single_source_shortest_path(curie))


@search_blueprint.route("/mappings/<source_curie>/<target_curie>")
def all_mappings(source_curie: str, target_curie: str):
    """Return all shortest paths of xrefs between the two identifiers."""
    if source_curie not in canonicalizer.graph:
        return jsonify(
            success=False,
            query=dict(source_curie=source_curie, target_curie=target_curie),
            message="could not find source curie",
        )
    if target_curie not in canonicalizer.graph:
        return jsonify(
            success=False,
            query=dict(source_curie=source_curie, target_curie=target_curie),
            message="could not find target curie",
        )

    return jsonify(_all_shortest_paths(source_curie, target_curie))


@search_blueprint.route("/mappings/summarize")
def summarize():
    """Summarize the mappings."""
    return summary_df.to_html(index=False)


@search_blueprint.route("/mappings/summarize_by/<prefix>")
def summarize_one(prefix: str):
    """Summarize the mappings."""
    norm_prefix = bioregistry.normalize_prefix(prefix)
    if norm_prefix is None:
        return abort(500, f"invalid prefix: {prefix}")
    in_df = summary_df.loc[summary_df[TARGET_PREFIX] == norm_prefix, [SOURCE_PREFIX, "count"]]
    out_df = summary_df.loc[summary_df[SOURCE_PREFIX] == norm_prefix, [TARGET_PREFIX, "count"]]
    return f"""
    <h1>Incoming Mappings to {norm_prefix}</h1>
    {in_df.to_html(index=False)}
    <h1>Outgoing Mappings from {norm_prefix}</h1>
    {out_df.to_html(index=False)}
    """


@search_blueprint.route("/canonicalize/<curie>")
def canonicalize(curie: str):
    """Return the best CURIE."""
    # TODO maybe normalize the curie first?
    norm_prefix, norm_identifier = normalize_curie(curie)
    if norm_prefix is None or norm_identifier is None:
        return jsonify(
            query=curie,
            normalizable=False,
        )

    norm_curie = f"{norm_prefix}:{norm_identifier}"

    rv: Dict[str, Any] = dict(query=curie)
    if norm_curie != curie:
        rv["norm_curie"] = norm_curie

    if norm_curie not in canonicalizer.graph:
        rv["found"] = False
    else:
        result_curie = canonicalizer.canonicalize(norm_curie)
        rv.update(
            found=True,
            result=result_curie,
            mappings=url_for(
                f".{all_mappings.__name__}",
                source_curie=norm_curie,
                target_curie=result_curie,
            ),
        )

    return jsonify(rv)


def get_app(paths: Union[None, str, Iterable[str]] = None) -> Flask:
    """Build the Flask app."""
    app = Flask(__name__)
    Swagger(app)

    logger.info("using bootstrap_flask %s", VERSION_BOOTSTRAP)
    Bootstrap(app)

    if paths is None:
        df = ensure_inspector_javert_df()
    elif isinstance(paths, str):
        df = pd.read_csv(paths, sep="\t", dtype=str)
    else:
        df = pd.concat(pd.read_csv(path, sep="\t", dtype=str) for path in paths)

    app.config["summary"] = summarize_xref_df(df)
    app.config["summary_provenances"] = summarize_xref_provenances_df(df)
    # TODO allow for specification of priorities in the canonicalizer
    app.config["canonicalizer"] = Canonicalizer.from_df(df)
    app.register_blueprint(search_blueprint)
    return app


def summarize_xref_df(df: pd.DataFrame) -> pd.DataFrame:
    """Get all meta-mappings."""
    return _summarize(df, [SOURCE_PREFIX, TARGET_PREFIX])


def summarize_xref_provenances_df(df: pd.DataFrame) -> pd.DataFrame:
    """Get all meta-mappings."""
    return _summarize(df, [SOURCE_PREFIX, TARGET_PREFIX, PROVENANCE])


def _summarize(df: pd.DataFrame, columns) -> pd.DataFrame:
    """Get all meta-mappings."""
    rv = df[columns].groupby(columns).size().reset_index()
    rv.columns = [*columns, "count"]
    rv.sort_values("count", inplace=True, ascending=False)
    return rv



# -*- coding: utf-8 -*-

"""Convert ICD11 to OBO.

Run with python -m pyobo.sources.icd11 -v
"""

import json
import logging
import os
from typing import Any, Iterable, Mapping, Set

import click
from more_click import verbose_option
from tqdm.auto import tqdm

from ..sources.icd_utils import (
    ICD11_TOP_LEVEL_URL,
    get_child_identifiers,
    get_icd,
    visiter,
)
from ..struct import Obo, Reference, Synonym, Term
from ..utils.path import prefix_directory_join

__all__ = [
    "ICD11Getter",
]

logger = logging.getLogger(__name__)

PREFIX = "icd11"


class ICD11Getter(Obo):
    """An ontology representation of ICD-11."""

    ontology = PREFIX
    dynamic_version = True

    def iter_terms(self, force: bool = False) -> Iterable[Term]:
        """Iterate over terms in the ontology."""
        return iterate_icd11()


def get_obo() -> Obo:
    """Get ICD11 as OBO."""
    return ICD11Getter()


def iterate_icd11() -> Iterable[Term]:
    """Iterate over the terms in ICD11.

    The API doesn't seem to have a rate limit, but returns pretty slow.
    This means that it only gets results at at about 5 calls/second.
    Get ready to be patient - the API token expires every hour so there's
    a caching mechanism with :mod:`cachier` that gets a new one every hour.
    """
    res = get_icd(ICD11_TOP_LEVEL_URL)
    res_json = res.json()

    version = res_json["releaseId"]
    directory = prefix_directory_join(PREFIX, version=version)

    with open(os.path.join(directory, "top.json"), "w") as file:
        json.dump(res_json, file, indent=2)

    tqdm.write(f'There are {len(res_json["child"])} top level entities')

    visited_identifiers: Set[str] = set()
    for identifier in get_child_identifiers(ICD11_TOP_LEVEL_URL, res_json):
        yield from visiter(
            identifier,
            visited_identifiers,
            directory,
            endpoint=ICD11_TOP_LEVEL_URL,
            converter=_extract_icd11,
        )


def _extract_icd11(res_json: Mapping[str, Any]) -> Term:
    identifier = res_json["@id"][len(ICD11_TOP_LEVEL_URL) :].lstrip("/")
    definition = res_json["definition"]["@value"] if "definition" in res_json else None
    name = res_json["title"]["@value"]
    synonyms = [Synonym(synonym["label"]["@value"]) for synonym in res_json.get("synonym", [])]
    parents = [
        Reference(prefix=PREFIX, identifier=url[len("http://id.who.int/icd/entity/") :])
        for url in res_json["parent"]
        if url[len("http://id.who.int/icd/entity/") :]
    ]
    return Term(
        reference=Reference(prefix=PREFIX, identifier=identifier, name=name),
        definition=definition,
        synonyms=synonyms,
        parents=parents,
    )


@click.command()
@verbose_option
def _main():
    get_obo().write_default(use_tqdm=True)


if __name__ == "__main__":
    _main()



# -*- coding: utf-8 -*-

"""Converter for FamPlex."""

import logging
from collections import defaultdict
from typing import Iterable, List, Mapping, Tuple

import bioregistry
from pystow.utils import get_commit

from pyobo import get_name_id_mapping
from pyobo.struct import Obo, Reference, Term
from pyobo.struct.typedef import has_member, has_part, is_a, part_of
from pyobo.utils.io import multidict
from pyobo.utils.path import ensure_df

logger = logging.getLogger(__name__)

PREFIX = "fplx"


class FamPlexGetter(Obo):
    """An ontology representation of FamPlex."""

    ontology = PREFIX
    dynamic_version = True
    typedefs = [has_member, has_part, is_a, part_of]

    def _get_version(self) -> str:
        return get_commit("sorgerlab", "famplex")

    def iter_terms(self, force: bool = False) -> Iterable[Term]:
        """Iterate over terms in the ontology."""
        return get_terms(force=force, version=self._version_or_raise)


def get_obo(force: bool = False) -> Obo:
    """Get FamPlex as OBO."""
    return FamPlexGetter(force=force)


def get_terms(version: str, force: bool = False) -> Iterable[Term]:
    """Get the FamPlex terms."""
    base_url = f"https://raw.githubusercontent.com/sorgerlab/famplex/{version}"

    entities_url = f"{base_url}/entities.csv"
    entities_df = ensure_df(PREFIX, url=entities_url, version=version, dtype=str, force=force)

    relations_url = f"{base_url}/relations.csv"
    relations_df = ensure_df(
        PREFIX, url=relations_url, version=version, header=None, sep=",", dtype=str, force=force
    )

    definitions_url = f"{base_url}/descriptions.csv"
    definitions_df = ensure_df(
        PREFIX,
        url=definitions_url,
        version=version,
        header=None,
        sep=",",
        dtype=str,
        force=force,
    )
    id_to_definition: Mapping[str, Tuple[str, str]] = {
        identifier: (definition, provenance)
        for identifier, provenance, definition in definitions_df.values
    }

    id_xrefs = _get_xref_df(version)

    hgnc_name_to_id = get_name_id_mapping("hgnc")
    in_edges = defaultdict(list)
    out_edges = defaultdict(list)
    for h_ns, h_name, r, t_ns, t_name in relations_df.values:
        if h_ns == "HGNC":
            h_identifier = hgnc_name_to_id.get(h_name)
            if h_identifier is None:
                logger.warning(
                    "[%s] could not look up HGNC identifier for gene: %s", PREFIX, h_name
                )
                continue
            h = Reference(prefix="hgnc", identifier=h_identifier, name=h_name)
        elif h_ns == "FPLX":
            h = Reference(prefix="fplx", identifier=h_name, name=h_name)
        elif h_ns == "UP":
            continue
        else:
            logger.exception(h_ns)
            raise
        if t_ns == "HGNC":
            t_identifier = hgnc_name_to_id.get(t_name)
            if t_identifier is None:
                logger.warning(
                    "[%s] could not look up HGNC identifier for gene: %s", PREFIX, t_name
                )
            t = Reference(prefix="hgnc", identifier=t_identifier, name=t_name)
        elif t_ns == "FPLX":
            t = Reference(prefix="fplx", identifier=t_name, name=t_name)
        elif h_ns == "UP":
            continue
        else:
            raise

        out_edges[h].append((r, t))
        in_edges[t].append((r, h))

    for (entity,) in entities_df.values:
        reference = Reference(prefix=PREFIX, identifier=entity, name=entity)
        definition, provenance = id_to_definition.get(entity, (None, None))
        provenance_reference = (
            Reference.from_curie(provenance) if isinstance(provenance, str) else None
        )
        term = Term(
            reference=reference,
            definition=definition,
            provenance=[] if provenance_reference is None else [provenance_reference],
        )

        for xref_reference in id_xrefs.get(entity, []):
            term.append_xref(xref_reference)

        for r, t in out_edges.get(reference, []):
            if r == "isa" and t.prefix == "fplx":
                term.append_parent(t)
            elif r == "isa":
                term.append_relationship(is_a, t)
            elif r == "partof":
                term.append_relationship(part_of, t)
            else:
                logging.warning("unhandled relation %s", r)

        for r, h in in_edges.get(reference, []):
            if r == "isa":
                term.append_relationship(has_member, h)
            elif r == "partof":
                term.append_relationship(has_part, h)
            else:
                logging.warning("unhandled relation %s", r)
        yield term


def _get_xref_df(version: str) -> Mapping[str, List[Reference]]:
    base_url = f"https://raw.githubusercontent.com/sorgerlab/famplex/{version}"
    xrefs_url = f"{base_url}/equivalences.csv"
    xrefs_df = ensure_df(PREFIX, url=xrefs_url, version=version, header=None, sep=",", dtype=str)

    # Normalize nextprot families
    ns_remapping = {
        "NXP": "nextprot.family",
    }
    xrefs_df[0] = xrefs_df[0].map(lambda s: ns_remapping.get(s, s))
    xrefs_df[1] = [
        bioregistry.standardize_identifier(xref_prefix, xref_identifier)
        if xref_prefix != "nextprot.family"
        else xref_identifier[len("FA:") :]
        for xref_prefix, xref_identifier in xrefs_df[[0, 1]].values
    ]

    xrefs_df[0] = xrefs_df[0].map(bioregistry.normalize_prefix)
    xrefs_df = xrefs_df[xrefs_df[0].notna()]
    xrefs_df = xrefs_df[xrefs_df[0] != "bel"]
    return multidict(
        (identifier, Reference(prefix=xref_prefix, identifier=xref_identifier))
        for xref_prefix, xref_identifier, identifier in xrefs_df.values
    )


if __name__ == "__main__":
    FamPlexGetter.cli()



# -*- coding: utf-8 -*-

"""Parser for the MeSH descriptors."""

import datetime
import itertools as itt
import logging
import re
from typing import Any, Dict, Iterable, List, Mapping, Optional, Set, Tuple
from xml.etree.ElementTree import Element

from tqdm.auto import tqdm

from pyobo.identifier_utils import standardize_ec
from pyobo.struct import Obo, Reference, Synonym, Term
from pyobo.utils.cache import cached_json, cached_mapping
from pyobo.utils.io import parse_xml_gz
from pyobo.utils.path import ensure_path, prefix_directory_join

__all__ = [
    "MeSHGetter",
]

logger = logging.getLogger(__name__)

PREFIX = "mesh"
NOW_YEAR = str(datetime.datetime.now().year)
CAS_RE = re.compile(r"^\d{1,7}\-\d{2}\-\d$")
UNII_RE = re.compile(r"[0-9A-Za-z]{10}$")


class MeSHGetter(Obo):
    """An ontology representation of the Medical Subject Headings."""

    ontology = bioversions_key = PREFIX

    def _get_version(self) -> Optional[str]:
        return NOW_YEAR

    def iter_terms(self, force: bool = False) -> Iterable[Term]:
        """Iterate over terms in the ontology."""
        return get_terms(version=self._version_or_raise, force=force)


def get_obo(force: bool = False) -> Obo:
    """Get MeSH as OBO."""
    return MeSHGetter(force=force)


def get_tree_to_mesh_id(version: str) -> Mapping[str, str]:
    """Get a mapping from MeSH tree numbers to their MeSH identifiers."""

    @cached_mapping(
        path=prefix_directory_join(PREFIX, name="mesh_tree.tsv", version=version),
        header=["mesh_tree_number", "mesh_id"],
    )
    def _inner():
        mesh = ensure_mesh_descriptors(version=version)
        rv = {}
        for entry in mesh:
            mesh_id = entry["identifier"]
            for tree_number in entry["tree_numbers"]:
                rv[tree_number] = mesh_id
        return rv

    return _inner()


def get_terms(version: str, force: bool = False) -> Iterable[Term]:
    """Get MeSH OBO terms."""
    mesh_id_to_term: Dict[str, Term] = {}

    descriptors = ensure_mesh_descriptors(version=version, force=force)
    supplemental_records = ensure_mesh_supplemental_records(version=version, force=force)

    for entry in itt.chain(descriptors, supplemental_records):
        identifier = entry["identifier"]
        name = entry["name"]
        definition = entry.get("scope_note")

        xrefs: List[Reference] = []
        synonyms: Set[str] = set()
        for concept in entry["concepts"]:
            synonyms.add(concept["name"])
            for term in concept["terms"]:
                synonyms.add(term["name"])
            for xref_prefix, xref_identifier in concept.get("xrefs", []):
                xrefs.append(Reference(prefix=xref_prefix, identifier=xref_identifier))

        mesh_id_to_term[identifier] = Term(
            definition=definition,
            reference=Reference(prefix=PREFIX, identifier=identifier, name=name),
            synonyms=[Synonym(name=synonym) for synonym in synonyms if synonym != name],
            xrefs=xrefs,
        )

    for entry in descriptors:
        mesh_id_to_term[entry["identifier"]].parents = [
            mesh_id_to_term[parent_descriptor_id].reference
            for parent_descriptor_id in entry["parents"]
        ]

    return mesh_id_to_term.values()


def ensure_mesh_descriptors(
    version: str, force: bool = False, force_process: bool = False
) -> List[Mapping[str, Any]]:
    """Get the parsed MeSH dictionary, and cache it if it wasn't already."""

    @cached_json(path=prefix_directory_join(PREFIX, name="desc.json", version=version), force=force)
    def _inner():
        path = ensure_path(PREFIX, url=get_descriptors_url(version), version=version)
        root = parse_xml_gz(path)
        return get_descriptor_records(root, id_key="DescriptorUI", name_key="DescriptorName/String")

    return _inner()


def get_descriptors_url(version: str) -> str:
    """Get the MeSH descriptors URL for the given version."""
    if version == NOW_YEAR:
        return f"https://nlmpubs.nlm.nih.gov/projects/mesh/MESH_FILES/xmlmesh/desc{version}.gz"
    return f"https://nlmpubs.nlm.nih.gov/projects/mesh/{version}/xmlmesh/desc{version}.gz"


def get_supplemental_url(version: str) -> str:
    """Get the MeSH supplemental URL for the given version."""
    if version == NOW_YEAR:
        return f"https://nlmpubs.nlm.nih.gov/projects/mesh/MESH_FILES/xmlmesh/supp{version}.gz"
    return f"https://nlmpubs.nlm.nih.gov/projects/mesh/{version}/xmlmesh/supp{version}.gz"


def ensure_mesh_supplemental_records(version: str, force: bool = False) -> List[Mapping[str, Any]]:
    """Get the parsed MeSH dictionary, and cache it if it wasn't already."""

    @cached_json(path=prefix_directory_join(PREFIX, name="supp.json", version=version), force=force)
    def _inner():
        path = ensure_path(PREFIX, url=get_supplemental_url(version), version=version)
        root = parse_xml_gz(path)
        return get_descriptor_records(
            root, id_key="SupplementalRecordUI", name_key="SupplementalRecordName/String"
        )

    return _inner()


def get_descriptor_records(element: Element, id_key: str, name_key) -> List[Dict[str, Any]]:
    """Get MeSH descriptor records."""
    logger.info("extract MeSH descriptors, concepts, and terms")

    rv: List[Dict[str, Any]] = [
        get_descriptor_record(descriptor, id_key=id_key, name_key=name_key)
        for descriptor in tqdm(element, desc="Getting MeSH Descriptors", unit_scale=True)
    ]
    logger.debug(f"got {len(rv)} descriptors")

    # cache tree numbers
    tree_number_to_descriptor_ui = {
        tree_number: descriptor["identifier"]
        for descriptor in rv
        for tree_number in descriptor["tree_numbers"]
    }
    logger.debug(f"got {len(tree_number_to_descriptor_ui)} tree mappings")

    # add in parents to each descriptor based on their tree numbers
    for descriptor in rv:
        parents_descriptor_uis = set()
        for tree_number in descriptor["tree_numbers"]:
            try:
                parent_tn, self_tn = tree_number.rsplit(".", 1)
            except ValueError:
                logger.debug("No dot for %s", tree_number)
                continue

            parent_descriptor_ui = tree_number_to_descriptor_ui.get(parent_tn)
            if parent_descriptor_ui is not None:
                parents_descriptor_uis.add(parent_descriptor_ui)
            else:
                logger.debug("missing tree number: %s", parent_tn)

        descriptor["parents"] = list(parents_descriptor_uis)

    return rv


def get_scope_note(descriptor_record) -> Optional[str]:
    """Get the scope note from the preferred concept in a term's record."""
    for concept in descriptor_record["concepts"]:
        scope_note = concept.get("ScopeNote")
        if scope_note is not None:
            return scope_note.replace("\\n", "\n").strip()
    return None


def get_descriptor_record(
    element: Element,
    id_key: str,
    name_key: str,
) -> Dict[str, Any]:
    """Get descriptor records from the main element.

    :param element: An XML element
    :param id_key: For descriptors, set to 'DescriptorUI'. For supplement, set to 'SupplementalRecordUI'
    :param name_key: For descriptors, set to 'DescriptorName/String'.
     For supplement, set to 'SupplementalRecordName/String'
    """
    concepts = get_concept_records(element)
    scope_note = get_scope_note(concepts)
    rv = {
        "identifier": element.findtext(id_key),
        "name": element.findtext(name_key),
        "tree_numbers": sorted(
            {x.text for x in element.findall("TreeNumberList/TreeNumber") if x.text}
        ),
        "concepts": concepts,
        # TODO handle AllowableQualifiersList
    }
    if scope_note:
        rv["scope_note"] = scope_note
    return rv


def get_concept_records(element: Element) -> List[Mapping[str, Any]]:
    """Get concepts from a record."""
    return [get_concept_record(e) for e in element.findall("ConceptList/Concept")]


def _get_xrefs(element: Element) -> List[Tuple[str, str]]:
    raw_registry_numbers: List[str] = sorted(
        {e.text for e in element.findall("RelatedRegistryNumberList/RegistryNumber") if e.text}
    )
    registry_number = element.findtext("RegistryNumber")
    if registry_number is not None:
        raw_registry_numbers.append(registry_number)
    raw_registry_numbers = [x for x in raw_registry_numbers if x != "0"]

    rv = []
    for registry_number in raw_registry_numbers:
        if registry_number == "0":
            continue
        elif registry_number.startswith("txid"):
            rv.append(("NCBITaxon", registry_number[4:]))
        elif registry_number.startswith("EC "):
            rv.append(("eccode", standardize_ec(registry_number[3:])))
        elif CAS_RE.fullmatch(registry_number):
            rv.append(("cas", registry_number))
        elif UNII_RE.fullmatch(registry_number):
            rv.append(("unii", registry_number))
        else:
            tqdm.write(f"Unhandled xref: {registry_number}")
    return rv


def get_concept_record(element: Element) -> Mapping[str, Any]:
    """Get a single MeSH concept record."""
    xrefs = _get_xrefs(element)

    scope_note = element.findtext("ScopeNote")
    if scope_note is not None:
        scope_note = scope_note.replace("\\n", "\n").strip()

    rv: Dict[str, Any] = {
        "concept_ui": element.findtext("ConceptUI"),
        "name": element.findtext("ConceptName/String"),
        "terms": get_term_records(element),
        # TODO handle ConceptRelationList
        **element.attrib,
    }
    semantic_types = sorted(
        {x.text for x in element.findall("SemanticTypeList/SemanticType/SemanticTypeUI") if x.text}
    )
    if semantic_types:
        rv["semantic_types"] = semantic_types
    if xrefs:
        rv["xrefs"] = xrefs
    if scope_note:
        rv["ScopeNote"] = scope_note
    return rv


def get_term_records(element: Element) -> List[Mapping[str, Any]]:
    """Get all of the terms for a concept."""
    return [get_term_record(term) for term in element.findall("TermList/Term")]


def get_term_record(element) -> Mapping[str, Any]:
    """Get a single MeSH term record."""
    return {
        "term_ui": element.findtext("TermUI"),
        "name": element.findtext("String"),
        **element.attrib,
    }


def _text_or_bust(element: Element, name: str) -> str:
    n = element.findtext(name)
    if n is None:
        raise ValueError
    return n


def _get_descriptor_qualifiers(descriptor: Element) -> List[Mapping[str, str]]:
    return [
        {
            "qualifier_ui": _text_or_bust(qualifier, "QualifierUI"),
            "name": _text_or_bust(qualifier, "QualifierName/String"),
        }
        for qualifier in descriptor.findall(
            "AllowableQualifiersList/AllowableQualifier/QualifierReferredTo"
        )
    ]


if __name__ == "__main__":
    get_obo(force=True).write_default(force=True, write_obo=True)



# -*- coding: utf-8 -*-

"""GMT utilities."""

from pathlib import Path
from typing import Iterable, Set, Tuple, Union

GMTSummary = Tuple[str, str, Set[str]]
WikiPathwaysGMTSummary = Tuple[str, str, str, str, str, Set[str]]


def parse_gmt_file(path: Union[str, Path]) -> Iterable[GMTSummary]:
    """Return file as list of pathway - gene sets (ENTREZ-identifiers).

    :param path: path to GMT file
    :return: line-based processed file
    """
    with open(path) as file:
        for line in file:
            yield _process_line(line)


def _process_line(line: str) -> Tuple[str, str, Set[str]]:
    """Return the pathway name, url, and gene sets associated.

    :param line: gmt file line
    :return: pathway name
    :return: pathway info url
    :return: genes set associated
    """
    name, info, *entries = (p.strip() for p in line.split("\t"))
    return name, info, set(entries)


def parse_wikipathways_gmt(path: Union[str, Path]) -> Iterable[WikiPathwaysGMTSummary]:
    """Parse WikiPathways GMT."""
    for name, info, entries in parse_gmt_file(path):
        name, version, identifier, species = name.split("%")
        version = version.split("_")[1]
        revision = info.rsplit("_", 1)[1].lstrip("r")
        yield identifier, version, revision, name, species, entries



# -*- coding: utf-8 -*-

"""Convert ICD-10 to OBO.

Run with python -m pyobo.sources.icd10 -v
"""

import logging
from typing import Any, Iterable, Mapping, Set

import click
from more_click import verbose_option
from tqdm.auto import tqdm

from ..sources.icd_utils import (
    ICD10_TOP_LEVEL_URL,
    get_child_identifiers,
    get_icd,
    visiter,
)
from ..struct import Obo, Reference, Synonym, Term
from ..utils.path import prefix_directory_join

__all__ = [
    "ICD10Getter",
]

logger = logging.getLogger(__name__)

PREFIX = "icd10"
VERSION = "2016"


class ICD10Getter(Obo):
    """An ontology representation of ICD-10."""

    ontology = PREFIX
    dynamic_version = True

    def iter_terms(self, force: bool = False) -> Iterable[Term]:
        """Iterate over terms in the ontology."""
        return iter_terms()


def get_obo() -> Obo:
    """Get ICD-10 as OBO."""
    return ICD10Getter()


def iter_terms() -> Iterable[Term]:
    """Iterate over ICD-10 terms."""
    r = get_icd(ICD10_TOP_LEVEL_URL)
    res_json = r.json()

    directory = prefix_directory_join(PREFIX, version=VERSION)

    chapter_urls = res_json["child"]
    tqdm.write(f"there are {len(chapter_urls)} chapters")

    visited_identifiers: Set[str] = set()
    for identifier in get_child_identifiers(ICD10_TOP_LEVEL_URL, res_json):
        yield from visiter(
            identifier,
            visited_identifiers,
            directory,
            endpoint=ICD10_TOP_LEVEL_URL,
            converter=_extract_icd10,
        )


def _extract_icd10(res_json: Mapping[str, Any]) -> Term:
    identifier = res_json["code"]
    name = res_json["title"]["@value"]
    synonyms = [Synonym(synonym["label"]["@value"]) for synonym in res_json.get("synonym", [])]
    parents = [
        Reference(prefix=PREFIX, identifier=url[len(ICD10_TOP_LEVEL_URL) :])
        for url in res_json["parent"]
        if url[len(ICD10_TOP_LEVEL_URL) :]
    ]
    rv = Term(
        reference=Reference(prefix=PREFIX, identifier=identifier, name=name),
        synonyms=synonyms,
        parents=parents,
    )

    rv.append_property("class_kind", res_json["classKind"])

    return rv


@click.command()
@verbose_option
def _main():
    get_obo().write_default(use_tqdm=True)


if __name__ == "__main__":
    _main()



# -*- coding: utf-8 -*-

"""Utilities for caching files from NDEx."""

import json
import os
from typing import Any, Iterable, List, Mapping, Tuple

import requests
from tqdm.auto import tqdm

from .path import prefix_directory_join

__all__ = [
    "CX",
    "ensure_ndex_network",
    "ensure_ndex_network_set",
    "iterate_aspect",
]

NDEX_BASE_URL = "http://public.ndexbio.org/v2"
NETWORK_ENDPOINT = f"{NDEX_BASE_URL}/network"
NETWORKSET_ENDPOINT = f"{NDEX_BASE_URL}/networkset"
CX = List[Mapping[str, Any]]


def iterate_aspect(cx: CX, aspect: str) -> Iterable[Any]:
    """Iterate over a given aspect."""
    for element in cx:
        if aspect in element:
            yield from element[aspect]


def ensure_ndex_network(prefix: str, uuid: str, force: bool = False) -> CX:
    """Ensure a network from NDEx is cached."""
    path = prefix_directory_join(prefix, "ndex", name=f"{uuid}.json")
    if os.path.exists(path) and not force:
        with open(path) as file:
            return json.load(file)

    res = requests.get(f"{NETWORK_ENDPOINT}/{uuid}")
    res_json = res.json()
    with open(path, "w") as file:
        json.dump(res_json, file, indent=2)
    return res_json


def ensure_ndex_network_set(
    prefix: str, uuid: str, use_tqdm: bool = False, force: bool = False
) -> Iterable[Tuple[str, CX]]:
    """Ensure the list of networks that goes with NCI PID on NDEx."""
    it = _help_ensure_ndex_network_set(prefix, uuid, force=force)
    if use_tqdm:
        it = tqdm(it, desc=f"Downloading ndex:{uuid}")
    for network_uuid in it:
        yield network_uuid, ensure_ndex_network(prefix, network_uuid, force=force)


def _help_ensure_ndex_network_set(prefix: str, uuid: str, force: bool = False) -> List[str]:
    """Ensure the list of networks that goes with NCI PID on NDEx."""
    networkset_path = prefix_directory_join(prefix, name="networks.txt")
    if os.path.exists(networkset_path) and not force:
        with open(networkset_path) as file:
            return sorted(line.strip() for line in file)

    url = f"{NETWORKSET_ENDPOINT}/{uuid}"
    res = requests.get(url)
    res_json = res.json()
    network_uuids = res_json["networks"]
    with open(networkset_path, "w") as file:
        for network_uuid in sorted(network_uuids):
            print(network_uuid, file=file)  # noqa: T201
    return network_uuids
